# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: signing.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rsigning.proto\x12\x1b\x61\x65\x61.open_aea.signing.v1_0_0"\xbc\x0c\n\x0eSigningMessage\x12O\n\x05\x65rror\x18\x05 \x01(\x0b\x32>.aea.open_aea.signing.v1_0_0.SigningMessage.Error_PerformativeH\x00\x12]\n\x0csign_message\x18\x06 \x01(\x0b\x32\x45.aea.open_aea.signing.v1_0_0.SigningMessage.Sign_Message_PerformativeH\x00\x12\x65\n\x10sign_transaction\x18\x07 \x01(\x0b\x32I.aea.open_aea.signing.v1_0_0.SigningMessage.Sign_Transaction_PerformativeH\x00\x12\x61\n\x0esigned_message\x18\x08 \x01(\x0b\x32G.aea.open_aea.signing.v1_0_0.SigningMessage.Signed_Message_PerformativeH\x00\x12i\n\x12signed_transaction\x18\t \x01(\x0b\x32K.aea.open_aea.signing.v1_0_0.SigningMessage.Signed_Transaction_PerformativeH\x00\x1a\xbd\x01\n\tErrorCode\x12W\n\nerror_code\x18\x01 \x01(\x0e\x32\x43.aea.open_aea.signing.v1_0_0.SigningMessage.ErrorCode.ErrorCodeEnum"W\n\rErrorCodeEnum\x12 \n\x1cUNSUCCESSFUL_MESSAGE_SIGNING\x10\x00\x12$\n UNSUCCESSFUL_TRANSACTION_SIGNING\x10\x01\x1a!\n\nRawMessage\x12\x13\n\x0braw_message\x18\x01 \x01(\x0c\x1a)\n\x0eRawTransaction\x12\x17\n\x0fraw_transaction\x18\x01 \x01(\x0c\x1a\'\n\rSignedMessage\x12\x16\n\x0esigned_message\x18\x01 \x01(\x0c\x1a/\n\x11SignedTransaction\x12\x1a\n\x12signed_transaction\x18\x01 \x01(\x0c\x1a\x16\n\x05Terms\x12\r\n\x05terms\x18\x01 \x01(\x0c\x1a\xb6\x01\n\x1dSign_Transaction_Performative\x12@\n\x05terms\x18\x01 \x01(\x0b\x32\x31.aea.open_aea.signing.v1_0_0.SigningMessage.Terms\x12S\n\x0fraw_transaction\x18\x02 \x01(\x0b\x32:.aea.open_aea.signing.v1_0_0.SigningMessage.RawTransaction\x1a\xaa\x01\n\x19Sign_Message_Performative\x12@\n\x05terms\x18\x01 \x01(\x0b\x32\x31.aea.open_aea.signing.v1_0_0.SigningMessage.Terms\x12K\n\x0braw_message\x18\x02 \x01(\x0b\x32\x36.aea.open_aea.signing.v1_0_0.SigningMessage.RawMessage\x1a|\n\x1fSigned_Transaction_Performative\x12Y\n\x12signed_transaction\x18\x01 \x01(\x0b\x32=.aea.open_aea.signing.v1_0_0.SigningMessage.SignedTransaction\x1ap\n\x1bSigned_Message_Performative\x12Q\n\x0esigned_message\x18\x01 \x01(\x0b\x32\x39.aea.open_aea.signing.v1_0_0.SigningMessage.SignedMessage\x1a_\n\x12\x45rror_Performative\x12I\n\nerror_code\x18\x01 \x01(\x0b\x32\x35.aea.open_aea.signing.v1_0_0.SigningMessage.ErrorCodeB\x0e\n\x0cperformativeb\x06proto3'
)


_SIGNINGMESSAGE = DESCRIPTOR.message_types_by_name["SigningMessage"]
_SIGNINGMESSAGE_ERRORCODE = _SIGNINGMESSAGE.nested_types_by_name["ErrorCode"]
_SIGNINGMESSAGE_RAWMESSAGE = _SIGNINGMESSAGE.nested_types_by_name["RawMessage"]
_SIGNINGMESSAGE_RAWTRANSACTION = _SIGNINGMESSAGE.nested_types_by_name["RawTransaction"]
_SIGNINGMESSAGE_SIGNEDMESSAGE = _SIGNINGMESSAGE.nested_types_by_name["SignedMessage"]
_SIGNINGMESSAGE_SIGNEDTRANSACTION = _SIGNINGMESSAGE.nested_types_by_name[
    "SignedTransaction"
]
_SIGNINGMESSAGE_TERMS = _SIGNINGMESSAGE.nested_types_by_name["Terms"]
_SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE = _SIGNINGMESSAGE.nested_types_by_name[
    "Sign_Transaction_Performative"
]
_SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE = _SIGNINGMESSAGE.nested_types_by_name[
    "Sign_Message_Performative"
]
_SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE = _SIGNINGMESSAGE.nested_types_by_name[
    "Signed_Transaction_Performative"
]
_SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE = _SIGNINGMESSAGE.nested_types_by_name[
    "Signed_Message_Performative"
]
_SIGNINGMESSAGE_ERROR_PERFORMATIVE = _SIGNINGMESSAGE.nested_types_by_name[
    "Error_Performative"
]
_SIGNINGMESSAGE_ERRORCODE_ERRORCODEENUM = _SIGNINGMESSAGE_ERRORCODE.enum_types_by_name[
    "ErrorCodeEnum"
]
SigningMessage = _reflection.GeneratedProtocolMessageType(
    "SigningMessage",
    (_message.Message,),
    {
        "ErrorCode": _reflection.GeneratedProtocolMessageType(
            "ErrorCode",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_ERRORCODE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.ErrorCode)
            },
        ),
        "RawMessage": _reflection.GeneratedProtocolMessageType(
            "RawMessage",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_RAWMESSAGE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.RawMessage)
            },
        ),
        "RawTransaction": _reflection.GeneratedProtocolMessageType(
            "RawTransaction",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_RAWTRANSACTION,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.RawTransaction)
            },
        ),
        "SignedMessage": _reflection.GeneratedProtocolMessageType(
            "SignedMessage",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGNEDMESSAGE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.SignedMessage)
            },
        ),
        "SignedTransaction": _reflection.GeneratedProtocolMessageType(
            "SignedTransaction",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGNEDTRANSACTION,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.SignedTransaction)
            },
        ),
        "Terms": _reflection.GeneratedProtocolMessageType(
            "Terms",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_TERMS,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.Terms)
            },
        ),
        "Sign_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Sign_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.Sign_Transaction_Performative)
            },
        ),
        "Sign_Message_Performative": _reflection.GeneratedProtocolMessageType(
            "Sign_Message_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.Sign_Message_Performative)
            },
        ),
        "Signed_Transaction_Performative": _reflection.GeneratedProtocolMessageType(
            "Signed_Transaction_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.Signed_Transaction_Performative)
            },
        ),
        "Signed_Message_Performative": _reflection.GeneratedProtocolMessageType(
            "Signed_Message_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.Signed_Message_Performative)
            },
        ),
        "Error_Performative": _reflection.GeneratedProtocolMessageType(
            "Error_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _SIGNINGMESSAGE_ERROR_PERFORMATIVE,
                "__module__": "signing_pb2"
                # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage.Error_Performative)
            },
        ),
        "DESCRIPTOR": _SIGNINGMESSAGE,
        "__module__": "signing_pb2"
        # @@protoc_insertion_point(class_scope:aea.open_aea.signing.v1_0_0.SigningMessage)
    },
)
_sym_db.RegisterMessage(SigningMessage)
_sym_db.RegisterMessage(SigningMessage.ErrorCode)
_sym_db.RegisterMessage(SigningMessage.RawMessage)
_sym_db.RegisterMessage(SigningMessage.RawTransaction)
_sym_db.RegisterMessage(SigningMessage.SignedMessage)
_sym_db.RegisterMessage(SigningMessage.SignedTransaction)
_sym_db.RegisterMessage(SigningMessage.Terms)
_sym_db.RegisterMessage(SigningMessage.Sign_Transaction_Performative)
_sym_db.RegisterMessage(SigningMessage.Sign_Message_Performative)
_sym_db.RegisterMessage(SigningMessage.Signed_Transaction_Performative)
_sym_db.RegisterMessage(SigningMessage.Signed_Message_Performative)
_sym_db.RegisterMessage(SigningMessage.Error_Performative)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _SIGNINGMESSAGE._serialized_start = 47
    _SIGNINGMESSAGE._serialized_end = 1643
    _SIGNINGMESSAGE_ERRORCODE._serialized_start = 551
    _SIGNINGMESSAGE_ERRORCODE._serialized_end = 740
    _SIGNINGMESSAGE_ERRORCODE_ERRORCODEENUM._serialized_start = 653
    _SIGNINGMESSAGE_ERRORCODE_ERRORCODEENUM._serialized_end = 740
    _SIGNINGMESSAGE_RAWMESSAGE._serialized_start = 742
    _SIGNINGMESSAGE_RAWMESSAGE._serialized_end = 775
    _SIGNINGMESSAGE_RAWTRANSACTION._serialized_start = 777
    _SIGNINGMESSAGE_RAWTRANSACTION._serialized_end = 818
    _SIGNINGMESSAGE_SIGNEDMESSAGE._serialized_start = 820
    _SIGNINGMESSAGE_SIGNEDMESSAGE._serialized_end = 859
    _SIGNINGMESSAGE_SIGNEDTRANSACTION._serialized_start = 861
    _SIGNINGMESSAGE_SIGNEDTRANSACTION._serialized_end = 908
    _SIGNINGMESSAGE_TERMS._serialized_start = 910
    _SIGNINGMESSAGE_TERMS._serialized_end = 932
    _SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE._serialized_start = 935
    _SIGNINGMESSAGE_SIGN_TRANSACTION_PERFORMATIVE._serialized_end = 1117
    _SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE._serialized_start = 1120
    _SIGNINGMESSAGE_SIGN_MESSAGE_PERFORMATIVE._serialized_end = 1290
    _SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE._serialized_start = 1292
    _SIGNINGMESSAGE_SIGNED_TRANSACTION_PERFORMATIVE._serialized_end = 1416
    _SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE._serialized_start = 1418
    _SIGNINGMESSAGE_SIGNED_MESSAGE_PERFORMATIVE._serialized_end = 1530
    _SIGNINGMESSAGE_ERROR_PERFORMATIVE._serialized_start = 1532
    _SIGNINGMESSAGE_ERROR_PERFORMATIVE._serialized_end = 1627
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test log parser."""

import re
import tempfile
from pathlib import Path
from typing import Tuple

from tests.test_autonomy.test_cli.base import BaseCliTest


TEST_LOGS = """[2022-04-14 17:45:43,962] [INFO] [agent] Entered in the 'registration_startup' round for period 0
[2022-04-14 17:44:43,323] [INFO] [agent] Entered in the 'registration_startup' behaviour state
[2022-04-14 17:45:41,128] [INFO] [agent] 'registration_startup' round is done with event: Event.FAST_FORWARD
[2022-04-14 17:45:41,128] [INFO] [agent] scheduling timeout of 7.0 seconds for event Event.ROUND_TIMEOUT with deadline 2022-04-14 17:45:46.527881
[2022-04-14 17:45:41,131] [INFO] [agent] Entered in the 'collect_observation' round for period 0
[2022-04-14 17:45:41,136] [INFO] [agent] Entered in the 'collect_observation' behaviour state
[2022-04-14 17:45:42,509] [INFO] [agent] 'collect_observation' round is done with event: Event.DONE
[2022-04-14 17:45:42,511] [INFO] [agent] scheduling timeout of 7.0 seconds for event Event.ROUND_TIMEOUT with deadline 2022-04-14 17:45:47.856572
"""

TEST_LOGS_WITH_MISSING_EVENTS = """
[2022-04-14 17:45:42,512] [INFO] [agent] Entered in the 'estimate_consensus' round for period 0
[2022-04-14 17:45:42,514] [INFO] [agent] Entered in the 'estimate' behaviour state
[2022-04-14 17:45:43,959] [INFO] [agent] 'estimate_consensus' round is done with event: Event.DONE
[2022-04-14 17:45:43,960] [INFO] [agent] scheduling timeout of 7.0 seconds for event Event.ROUND_TIMEOUT with deadline 2022-04-14 17:45:49.269590
[2022-04-14 17:45:43,962] [INFO] [agent] Entered in the 'tx_hash' round for period 0
[2022-04-14 17:45:43,968] [INFO] [agent] Entered in the 'tx_hash' behaviour state
[2022-04-14 17:45:46,927] [INFO] [agent] scheduling timeout of 7.0 seconds for event Event.ROUND_TIMEOUT with deadline 2022-04-14 17:45:52.089382
"""


EXPECTED_OUTPUT = (
    "agent         period: 0     round: registration_startup                      event: Event.FAST_FORWARD\n"
    "agent         period: 0     round: collect_observation                       event: Event.DONE\n"
    "\n"
    "ERRORS:\n"
)

EXPECTED_OUTPUT_WITH_MISSING_EVENTS = (
    "agent         period: 0     round: estimate_consensus                        event: Event.DONE\n"
    "agent         period: 0     round: tx_hash                                   event: N/A\n"
    "\n"
    "ERRORS:\n"
)


ANSI_COLOR_CHARACTERS_REGEX = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


class TestLogParser(BaseCliTest):
    """Test log parser."""

    cli_options: Tuple[str, ...] = ("analyse", "abci", "logs")

    def _check_output(self, logs: str, expected_output: str) -> None:
        """Check outputs."""

        with tempfile.TemporaryDirectory() as temp:
            log_file = Path(temp) / "log.txt"
            log_file.write_text(logs, encoding="utf-8")
            result = self.run_cli((str(log_file),))
            output = ANSI_COLOR_CHARACTERS_REGEX.sub("", result.stdout)
            assert output == expected_output

    def test_parser(
        self,
    ) -> None:
        """Test log parser."""

        self._check_output(TEST_LOGS, EXPECTED_OUTPUT)

    def test_parser_with_missing_events(
        self,
    ) -> None:
        """Test log parser."""

        self._check_output(
            TEST_LOGS_WITH_MISSING_EVENTS, EXPECTED_OUTPUT_WITH_MISSING_EVENTS
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""
Parses agent logs and outputs information about rounds and events, as well as errors.

Can be used both for e2e logs from tests as well as for individual agent logs.

./scripts/parse_logs.py -i log.txt
"""

import argparse
import re
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Dict


LOG_FILE_PATH = "/tmp/log"  # nosec

PERIOD_REGEX = r".*\[(?P<agent_id>.*)\] Entered in the '(?P<round_id>.*)' round for period (?P<period>\d+)"
EVENT_REGEX = r".*\[(?P<agent_id>.*)\]\s'(?P<round_id>.*)' round is done with event: (?P<event_id>.*)"
ERROR_REGEX = r".*E\s+AssertionError:\s(?P<message>.*)"


class Color(Enum):
    """Standard terminal color codes."""

    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    RESET = "\033[39m"


AGENT_COLORS = [
    Color.BLUE.value,
    Color.MAGENTA.value,
    Color.CYAN.value,
    Color.WHITE.value,
]


def parse_file(  # pylint: disable=too-many-locals
    filename: str = LOG_FILE_PATH,
) -> None:
    """Parses a log file, e2e or single agent."""
    periods: Dict[str, Any] = {}
    errors = []
    _id = 0
    with open(filename, "r", encoding="utf-8") as log:
        for line in log.readlines():

            m = re.match(PERIOD_REGEX, line)
            if m:
                agent_id = m.groupdict()["agent_id"]
                if agent_id not in periods.keys():
                    periods[agent_id] = []
                periods[agent_id].append(m.groupdict())
                continue

            m = re.match(EVENT_REGEX, line)
            if m:
                agent_id = m.groupdict()["agent_id"]
                if agent_id not in periods.keys():
                    periods[agent_id] = []
                periods[agent_id].append(m.groupdict())
                continue

            m = re.match(ERROR_REGEX, line, re.DOTALL)
            if m:
                errors.append(m.groupdict())
                continue

    for agent_id, period_list in periods.items():
        try:
            num_id = int(agent_id.replace("agent_", ""))
        except Exception:  # pylint: disable=broad-except
            num_id = _id
            _id += 1
        agent_color = AGENT_COLORS[num_id]
        for i, period in enumerate(period_list):

            # Check that this is a new period message
            if "period" not in period.keys():
                continue

            # Check that the next message is an event one
            event: str = ""
            if i < len(period_list) - 1 and "event_id" in period_list[i + 1].keys():
                event = period_list[i + 1]["event_id"]
                color: str = Color.GREEN.value if "DONE" in event else Color.RED.value
                event = color + event + Color.RESET.value
            else:
                event = Color.YELLOW.value + "N/A" + Color.RESET.value

            # Print the period info
            print(
                f"{agent_color}{period['agent_id']:<12}{Color.RESET.value}  period: {period['period']:<4}  round: {period['round_id']:<40}  event: {event}"
            )

    print("\nERRORS:")
    for error in errors:
        print(Color.RED.value + error["message"] + Color.RESET.value)


def parse_args() -> argparse.Namespace:
    """Get cli args."""
    script_name = Path(__file__).name
    parser = argparse.ArgumentParser(
        script_name,
        description=f"Extracts state transitions from agent logs:\n"
        f"./{script_name} -i abci0.txt",
    )
    required = parser.add_argument_group("required arguments")
    required.add_argument(
        "-i",
        "--infile",
        type=str,
        help="Input file name.",
    )
    arguments_ = parser.parse_args()
    return arguments_


def main(args: argparse.Namespace) -> None:
    """Main function."""
    if args.infile is None:
        print("Please provide path to abci log.")
        sys.exit(1)
    parse_file(args.infile)


if __name__ == "__main__":
    main(args=parse_args())



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the base round classes."""

import re
from copy import deepcopy
from enum import Enum
from typing import (
    Any,
    Callable,
    FrozenSet,
    Generator,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    cast,
)
from unittest import mock

import pytest

from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciAppDB,
    AbstractRound,
    BaseSynchronizedData,
    BaseTxPayload,
    CollectDifferentUntilAllRound,
    CollectDifferentUntilThresholdRound,
    CollectNonEmptyUntilThresholdRound,
    CollectSameUntilAllRound,
    CollectSameUntilThresholdRound,
    CollectionRound,
    ConsensusParams,
    OnlyKeeperSendsRound,
    ROUND_COUNT_DEFAULT,
    TransactionNotValidError,
    VotingRound,
)


MAX_PARTICIPANTS: int = 4


def get_participants() -> FrozenSet[str]:
    """Participants"""
    return frozenset([f"agent_{i}" for i in range(MAX_PARTICIPANTS)])


class DummyTxPayload(BaseTxPayload):
    """Dummy Transaction Payload."""

    transaction_type = "DummyPayload"
    _value: Optional[str]
    _vote: Optional[bool]

    def __init__(
        self,
        sender: str,
        value: Any,
        vote: Optional[bool] = False,
        round_count: int = ROUND_COUNT_DEFAULT,
    ) -> None:
        """Initialize a dummy transaction payload."""

        super().__init__(sender, None)
        self._value = value
        self._vote = vote
        self._round_count = round_count

    @property
    def value(self) -> Any:
        """Get the tx value."""
        return self._value

    @property
    def vote(self) -> Optional[bool]:
        """Get the vote value."""
        return self._vote


class DummySynchronizedSata(BaseSynchronizedData):
    """Dummy synchronized data for tests."""

    @property
    def most_voted_keeper_address(
        self,
    ) -> str:
        """Returns value for _most_voted_keeper_address."""
        return self.db.get_strict("most_voted_keeper_address")


def get_dummy_tx_payloads(
    participants: FrozenSet[str],
    value: Any = None,
    vote: Optional[bool] = False,
    is_value_none: bool = False,
) -> List[DummyTxPayload]:
    """Returns a list of DummyTxPayload objects."""
    return [
        DummyTxPayload(
            sender=agent,
            value=(value or agent) if not is_value_none else value,
            vote=vote,
        )
        for agent in sorted(participants)
    ]


class DummyRound(AbstractRound):
    """Dummy round."""

    round_id = "round_id"
    allowed_tx_type = DummyTxPayload.transaction_type
    payload_attribute = "value"

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """end_block method."""


class DummyCollectionRound(CollectionRound, DummyRound):
    """Dummy Class for CollectionRound"""


class DummyCollectDifferentUntilAllRound(CollectDifferentUntilAllRound, DummyRound):
    """Dummy Class for CollectDifferentUntilAllRound"""


class DummyCollectSameUntilAllRound(CollectSameUntilAllRound, DummyRound):
    """Dummy Class for CollectSameUntilThresholdRound"""


class DummyCollectDifferentUntilThresholdRound(
    CollectDifferentUntilThresholdRound, DummyRound
):
    """Dummy Class for CollectDifferentUntilThresholdRound"""


class DummyCollectSameUntilThresholdRound(CollectSameUntilThresholdRound, DummyRound):
    """Dummy Class for CollectSameUntilThresholdRound"""


class DummyOnlyKeeperSendsRound(OnlyKeeperSendsRound, DummyRound):
    """Dummy Class for OnlyKeeperSendsRound"""

    fail_event = "FAIL_EVENT"


class DummyVotingRound(VotingRound, DummyRound):
    """Dummy Class for VotingRound"""


class DummyCollectNonEmptyUntilThresholdRound(
    CollectNonEmptyUntilThresholdRound, DummyRound
):
    """Dummy Class for `CollectNonEmptyUntilThresholdRound`"""


class BaseRoundTestClass:
    """Base test class."""

    synchronized_data: BaseSynchronizedData
    participants: FrozenSet[str]
    consensus_params: ConsensusParams

    _synchronized_data_class: Type[BaseSynchronizedData]
    _event_class: Any

    @classmethod
    def setup(
        cls,
    ) -> None:
        """Setup test class."""

        cls.participants = get_participants()
        cls.synchronized_data = cls._synchronized_data_class(
            db=AbciAppDB(
                setup_data=dict(
                    participants=[cls.participants], all_participants=[cls.participants]
                ),
            )
        )  # type: ignore
        cls.consensus_params = ConsensusParams(max_participants=MAX_PARTICIPANTS)

    def _test_no_majority_event(self, round_obj: AbstractRound) -> None:
        """Test the NO_MAJORITY event."""
        with mock.patch.object(round_obj, "is_majority_possible", return_value=False):
            result = round_obj.end_block()
            assert result is not None
            _, event = result
            assert event == self._event_class.NO_MAJORITY

    def _complete_run(self, test_runner: Generator, iter_count: int = 4) -> None:
        """
        This method represents logic to execute test logic defined in _test_round method.

        _test_round should follow these steps

        1. process first payload
        2. yield test_round
        3. test collection, end_block and thresholds
        4. process rest of the payloads
        5. yield test_round
        6. yield synchronized_data, event ( returned from end_block )
        7. test synchronized_data and event

        :param test_runner: test runner
        :param iter_count: iter_count
        """

        for _ in range(iter_count):
            next(test_runner)


class BaseCollectDifferentUntilAllRoundTest(BaseRoundTestClass):
    """Tests for rounds derived from CollectDifferentUntilAllRound."""

    def _test_round(
        self,
        test_round: CollectDifferentUntilAllRound,
        round_payloads: List[BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test round."""

        first_payload = round_payloads.pop(0)
        test_round.process_payload(first_payload)

        yield test_round
        assert test_round.collection[first_payload.sender] == first_payload
        assert not test_round.collection_threshold_reached
        assert test_round.end_block() is None

        for payload in round_payloads:
            test_round.process_payload(payload)
        yield test_round
        assert test_round.collection_threshold_reached

        actual_next_synchronized_data = cast(
            self._synchronized_data_class,  # type: ignore
            synchronized_data_update_fn(deepcopy(self.synchronized_data), test_round),  # type: ignore
        )

        res = test_round.end_block()
        yield res
        if exit_event is None:
            assert res is exit_event
        else:
            assert res is not None
            synchronized_data, event = res
            synchronized_data = cast(self._synchronized_data_class, synchronized_data)  # type: ignore
            for behaviour_attr_getter in synchronized_data_attr_checks:
                assert behaviour_attr_getter(
                    synchronized_data
                ) == behaviour_attr_getter(actual_next_synchronized_data)
            assert event == exit_event
        yield


class BaseCollectSameUntilAllRoundTest(BaseRoundTestClass):
    """Tests for rounds derived from CollectSameUntilAllRound."""

    def _test_round(
        self,
        test_round: CollectSameUntilAllRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        most_voted_payload: Any,
        exit_event: Any,
        finished: bool,
    ) -> Generator:
        """Test rounds derived from CollectionRound."""

        (_, first_payload), *payloads = round_payloads.items()

        test_round.process_payload(first_payload)
        yield test_round
        assert test_round.collection[first_payload.sender] == first_payload
        assert not test_round.collection_threshold_reached
        assert test_round.end_block() is None

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: 1 votes are not enough for `CollectSameUntilAllRound`. "
            "Expected: `n_votes = max_participants = 4`",
        ):
            _ = test_round.common_payload

        for _, payload in payloads:
            test_round.process_payload(payload)
        yield test_round
        if finished:
            assert test_round.collection_threshold_reached
        assert test_round.common_payload == most_voted_payload

        actual_next_synchronized_data = cast(
            self._synchronized_data_class,  # type: ignore
            synchronized_data_update_fn(deepcopy(self.synchronized_data), test_round),  # type: ignore
        )
        res = test_round.end_block()
        yield res
        assert res is not None

        synchronized_data, event = res
        synchronized_data = cast(self._synchronized_data_class, synchronized_data)  # type: ignore

        for behaviour_attr_getter in synchronized_data_attr_checks:
            assert behaviour_attr_getter(synchronized_data) == behaviour_attr_getter(
                actual_next_synchronized_data
            )
        assert event == exit_event
        yield


class BaseCollectSameUntilThresholdRoundTest(BaseRoundTestClass):
    """Tests for rounds derived from CollectSameUntilThresholdRound."""

    def _test_round(
        self,
        test_round: CollectSameUntilThresholdRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        most_voted_payload: Any,
        exit_event: Any,
    ) -> Generator:
        """Test rounds derived from CollectionRound."""

        (_, first_payload), *payloads = round_payloads.items()

        test_round.process_payload(first_payload)
        yield test_round
        assert test_round.collection[first_payload.sender] == first_payload
        assert not test_round.threshold_reached
        assert test_round.end_block() is None

        self._test_no_majority_event(test_round)
        with pytest.raises(ABCIAppInternalError, match="not enough votes"):
            _ = test_round.most_voted_payload

        for _, payload in payloads:
            test_round.process_payload(payload)
        yield test_round
        assert test_round.threshold_reached
        assert test_round.most_voted_payload == most_voted_payload

        actual_next_synchronized_data = cast(
            self._synchronized_data_class,  # type: ignore
            synchronized_data_update_fn(deepcopy(self.synchronized_data), test_round),  # type: ignore
        )
        res = test_round.end_block()
        yield res
        assert res is not None

        synchronized_data, event = res
        synchronized_data = cast(self._synchronized_data_class, synchronized_data)  # type: ignore

        for behaviour_attr_getter in synchronized_data_attr_checks:
            assert behaviour_attr_getter(synchronized_data) == behaviour_attr_getter(
                actual_next_synchronized_data
            )
        assert event == exit_event
        yield


class BaseOnlyKeeperSendsRoundTest(BaseRoundTestClass):
    """Tests for rounds derived from OnlyKeeperSendsRound."""

    def _test_round(
        self,
        test_round: OnlyKeeperSendsRound,
        keeper_payloads: BaseTxPayload,
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test for rounds derived from OnlyKeeperSendsRound."""

        assert test_round.end_block() is None
        assert not test_round.has_keeper_sent_payload

        test_round.process_payload(keeper_payloads)
        yield test_round
        assert test_round.has_keeper_sent_payload

        yield test_round
        actual_next_synchronized_data = cast(
            self._synchronized_data_class,  # type: ignore
            synchronized_data_update_fn(deepcopy(self.synchronized_data), test_round),  # type: ignore
        )
        res = test_round.end_block()
        yield res
        assert res is not None

        synchronized_data, event = res
        synchronized_data = cast(self._synchronized_data_class, synchronized_data)  # type: ignore
        for behaviour_attr_getter in synchronized_data_attr_checks:
            assert behaviour_attr_getter(synchronized_data) == behaviour_attr_getter(
                actual_next_synchronized_data
            )
        assert event == exit_event
        yield


class BaseVotingRoundTest(BaseRoundTestClass):
    """Tests for rounds derived from VotingRound."""

    def _test_round(
        self,
        test_round: VotingRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
        threshold_check: Callable,
    ) -> Generator:
        """Test for rounds derived from VotingRound."""

        (sender, first_payload), *payloads = round_payloads.items()

        test_round.process_payload(first_payload)
        yield test_round
        assert not threshold_check(test_round)  # negative_vote_threshold_reached
        assert test_round.end_block() is None
        self._test_no_majority_event(test_round)

        for _, payload in payloads:
            test_round.process_payload(payload)
        yield test_round
        assert threshold_check(test_round)

        actual_next_synchronized_data = cast(
            self._synchronized_data_class,  # type: ignore
            synchronized_data_update_fn(deepcopy(self.synchronized_data), test_round),  # type: ignore
        )
        res = test_round.end_block()
        yield res
        assert res is not None

        synchronized_data, event = res
        synchronized_data = cast(self._synchronized_data_class, synchronized_data)  # type: ignore
        for behaviour_attr_getter in synchronized_data_attr_checks:
            assert behaviour_attr_getter(synchronized_data) == behaviour_attr_getter(
                actual_next_synchronized_data
            )
        assert event == exit_event
        yield

    def _test_voting_round_positive(
        self,
        test_round: VotingRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test for rounds derived from VotingRound."""

        return self._test_round(
            test_round,
            round_payloads,
            synchronized_data_update_fn,
            synchronized_data_attr_checks,
            exit_event,
            threshold_check=lambda x: x.positive_vote_threshold_reached,
        )

    def _test_voting_round_negative(
        self,
        test_round: VotingRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test for rounds derived from VotingRound."""

        return self._test_round(
            test_round,
            round_payloads,
            synchronized_data_update_fn,
            synchronized_data_attr_checks,
            exit_event,
            threshold_check=lambda x: x.negative_vote_threshold_reached,
        )

    def _test_voting_round_none(
        self,
        test_round: VotingRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test for rounds derived from VotingRound."""

        return self._test_round(
            test_round,
            round_payloads,
            synchronized_data_update_fn,
            synchronized_data_attr_checks,
            exit_event,
            threshold_check=lambda x: x.none_vote_threshold_reached,
        )


class BaseCollectDifferentUntilThresholdRoundTest(BaseRoundTestClass):
    """Tests for rounds derived from CollectDifferentUntilThresholdRound."""

    def _test_round(
        self,
        test_round: CollectDifferentUntilThresholdRound,
        round_payloads: Mapping[str, BaseTxPayload],
        synchronized_data_update_fn: Callable,
        synchronized_data_attr_checks: List[Callable],
        exit_event: Any,
    ) -> Generator:
        """Test for rounds derived from CollectDifferentUntilThresholdRound."""

        (_, first_payload), *payloads = round_payloads.items()

        test_round.process_payload(first_payload)
        yield test_round
        assert not test_round.collection_threshold_reached
        assert test_round.end_block() is None

        for _, payload in payloads:
            test_round.process_payload(payload)
        yield test_round
        assert test_round.collection_threshold_reached

        actual_next_synchronized_data = cast(
            self._synchronized_data_class,  # type: ignore
            synchronized_data_update_fn(deepcopy(self.synchronized_data), test_round),  # type: ignore
        )
        res = test_round.end_block()
        yield res
        assert res is not None

        synchronized_data, event = res
        synchronized_data = cast(self._synchronized_data_class, synchronized_data)  # type: ignore

        for behaviour_attr_getter in synchronized_data_attr_checks:
            assert behaviour_attr_getter(synchronized_data) == behaviour_attr_getter(
                actual_next_synchronized_data
            )
        assert event == exit_event
        yield


class BaseCollectNonEmptyUntilThresholdRound(
    BaseCollectDifferentUntilThresholdRoundTest
):
    """Tests for rounds derived from `CollectNonEmptyUntilThresholdRound`."""


class _BaseRoundTestClass(BaseRoundTestClass):
    """Base test class."""

    synchronized_data: BaseSynchronizedData
    participants: FrozenSet[str]
    consensus_params: ConsensusParams
    tx_payloads: List[DummyTxPayload]

    _synchronized_data_class = DummySynchronizedSata

    @classmethod
    def setup(
        cls,
    ) -> None:
        """Setup test class."""

        super().setup()
        cls.tx_payloads = get_dummy_tx_payloads(cls.participants)

    def _test_payload_with_wrong_round_count(
        self, test_round: AbstractRound, value: Optional[Any] = None
    ) -> None:
        """Test errors raised by pyaloads with wrong round count."""
        payload_with_wrong_round_count = DummyTxPayload("sender", value, False, 0)
        with pytest.raises(
            TransactionNotValidError,
            match=re.escape("Expected round count -1 and got 0."),
        ):
            test_round.check_payload(payload=payload_with_wrong_round_count)

        with pytest.raises(
            ABCIAppInternalError,
            match=re.escape("Expected round count -1 and got 0."),
        ):
            test_round.process_payload(payload=payload_with_wrong_round_count)


class TestCollectionRound(_BaseRoundTestClass):
    """Test class for CollectionRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = DummyCollectionRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        first_payload, *_ = self.tx_payloads
        test_round.process_payload(first_payload)
        assert test_round.collection[first_payload.sender] == first_payload

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: sender agent_0 has already sent value for round: round_id",
        ):
            test_round.process_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match=re.escape(
                "internal error: sender not in list of participants: ['agent_0', 'agent_1', 'agent_2', 'agent_3']"
            ),
        ):
            test_round.process_payload(DummyTxPayload("sender", "value"))

        with pytest.raises(
            TransactionNotValidError,
            match="sender agent_0 has already sent value for round: round_id",
        ):
            test_round.check_payload(first_payload)

        with pytest.raises(
            TransactionNotValidError,
            match=re.escape(
                "sender not in list of participants: ['agent_0', 'agent_1', 'agent_2', 'agent_3']"
            ),
        ):
            test_round.check_payload(DummyTxPayload("sender", "value"))

        self._test_payload_with_wrong_round_count(test_round)


class TestCollectDifferentUntilAllRound(_BaseRoundTestClass):
    """Test class for CollectDifferentUntilAllRound."""

    def test_run(
        self,
    ) -> None:
        """Run Tests."""

        test_round = DummyCollectDifferentUntilAllRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        first_payload, *payloads = self.tx_payloads
        test_round.process_payload(first_payload)
        assert not test_round.collection_threshold_reached

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: sender agent_0 has already sent value for round: round_id",
        ):
            test_round.process_payload(first_payload)

        with pytest.raises(
            TransactionNotValidError,
            match="sender agent_0 has already sent value for round: round_id",
        ):
            test_round.check_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: `CollectDifferentUntilAllRound` encountered a value 'agent_0' that already exists.",
        ):
            first_payload.sender = "other"
            test_round.process_payload(first_payload)

        with pytest.raises(
            TransactionNotValidError,
            match="`CollectDifferentUntilAllRound` encountered a value 'agent_0' that already exists.",
        ):
            test_round.check_payload(first_payload)

        for payload in payloads:
            assert not test_round.collection_threshold_reached
            test_round.process_payload(payload)

        assert test_round.collection_threshold_reached
        self._test_payload_with_wrong_round_count(test_round)


class TestCollectSameUntilAllRound(_BaseRoundTestClass):
    """Test class for CollectSameUntilAllRound."""

    def test_run(
        self,
    ) -> None:
        """Run Tests."""

        test_round = DummyCollectSameUntilAllRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        first_payload, *payloads = [
            DummyTxPayload(
                sender=agent,
                value="test",
            )
            for agent in sorted(self.participants)
        ]
        test_round.process_payload(first_payload)
        assert not test_round.collection_threshold_reached

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: sender agent_0 has already sent value for round: round_id",
        ):
            test_round.process_payload(first_payload)

        with pytest.raises(
            TransactionNotValidError,
            match="sender agent_0 has already sent value for round: round_id",
        ):
            test_round.check_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: `CollectSameUntilAllRound` encountered a value 'other' "
            "which is not the same as the already existing one: 'test",
        ):
            bad_payload = DummyTxPayload(
                sender="other",
                value="other",
            )
            test_round.process_payload(bad_payload)

        with pytest.raises(
            TransactionNotValidError,
            match="`CollectSameUntilAllRound` encountered a value 'other' "
            "which is not the same as the already existing one: 'test",
        ):
            test_round.check_payload(bad_payload)

        for payload in payloads:
            assert not test_round.collection_threshold_reached
            test_round.process_payload(payload)

        assert test_round.collection_threshold_reached
        self._test_payload_with_wrong_round_count(test_round, "test")


class TestCollectSameUntilThresholdRound(_BaseRoundTestClass):
    """Test CollectSameUntilThresholdRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = DummyCollectSameUntilThresholdRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        first_payload, *payloads = get_dummy_tx_payloads(
            self.participants, value="vote"
        )
        test_round.process_payload(first_payload)

        assert not test_round.threshold_reached
        with pytest.raises(ABCIAppInternalError, match="not enough votes"):
            _ = test_round.most_voted_payload

        for payload in payloads:
            test_round.process_payload(payload)

        assert test_round.threshold_reached
        assert test_round.most_voted_payload == "vote"

        self._test_payload_with_wrong_round_count(test_round)

    def test_run_with_none(
        self,
    ) -> None:
        """Run tests."""

        test_round = DummyCollectSameUntilThresholdRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        first_payload, *payloads = get_dummy_tx_payloads(
            self.participants,
            value=None,
            is_value_none=True,
        )
        test_round.process_payload(first_payload)

        assert not test_round.threshold_reached
        with pytest.raises(ABCIAppInternalError, match="not enough votes"):
            _ = test_round.most_voted_payload

        for payload in payloads:
            test_round.process_payload(payload)

        assert test_round.threshold_reached
        assert test_round.most_voted_payload is None


class TestOnlyKeeperSendsRound(_BaseRoundTestClass, BaseOnlyKeeperSendsRoundTest):
    """Test OnlyKeeperSendsRound."""

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = DummyOnlyKeeperSendsRound(
            synchronized_data=self.synchronized_data.update(
                most_voted_keeper_address="agent_0"
            ),
            consensus_params=self.consensus_params,
        )

        assert not test_round.has_keeper_sent_payload
        first_payload, *_ = self.tx_payloads
        test_round.process_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match="internal error: keeper already set the payload.",
        ):
            test_round.process_payload(first_payload)

        with pytest.raises(
            ABCIAppInternalError,
            match=re.escape(
                "internal error: sender not in list of participants: ['agent_0', 'agent_1', 'agent_2', 'agent_3']"
            ),
        ):
            test_round.process_payload(DummyTxPayload(sender="sender", value="sender"))

        with pytest.raises(
            ABCIAppInternalError, match="internal error: agent_1 not elected as keeper."
        ):
            test_round.process_payload(DummyTxPayload(sender="agent_1", value="sender"))

        with pytest.raises(
            TransactionNotValidError, match="keeper payload value already set."
        ):
            test_round.check_payload(first_payload)

        with pytest.raises(
            TransactionNotValidError,
            match=re.escape(
                "sender not in list of participants: ['agent_0', 'agent_1', 'agent_2', 'agent_3']"
            ),
        ):
            test_round.check_payload(DummyTxPayload(sender="sender", value="sender"))

        with pytest.raises(
            TransactionNotValidError, match="agent_1 not elected as keeper."
        ):
            test_round.check_payload(DummyTxPayload(sender="agent_1", value="sender"))

        self._test_payload_with_wrong_round_count(test_round)

    def test_keeper_payload_is_none(
        self,
    ) -> None:
        """Test keeper payload valur set to none."""

        keeper = "agent_0"
        self._complete_run(
            self._test_round(
                test_round=DummyOnlyKeeperSendsRound(
                    synchronized_data=self.synchronized_data.update(
                        most_voted_keeper_address=keeper,
                    ),
                    consensus_params=self.consensus_params,
                ),
                keeper_payloads=DummyTxPayload(keeper, None),
                synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data,
                synchronized_data_attr_checks=[],
                exit_event="FAIL_EVENT",
            )
        )


class TestVotingRound(_BaseRoundTestClass):
    """Test VotingRound."""

    def setup_test_voting_round(self) -> DummyVotingRound:
        """Setup test voting round"""
        return DummyVotingRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

    def test_vote_count(self) -> None:
        """Testing agent vote count"""
        test_round = self.setup_test_voting_round()
        a, b, c, d = self.participants
        for agents, vote in [((a, d), True), ((c,), False), ((b,), None)]:
            for payload in get_dummy_tx_payloads(frozenset(agents), vote=vote):
                test_round.process_payload(payload)
        assert dict(test_round.vote_count) == {True: 2, False: 1, None: 1}

        self._test_payload_with_wrong_round_count(test_round)

    def test_negative_threshold(
        self,
    ) -> None:
        """Runs test."""

        test_round = self.setup_test_voting_round()
        first_payload, *payloads = get_dummy_tx_payloads(self.participants, vote=False)
        test_round.process_payload(first_payload)

        assert not test_round.negative_vote_threshold_reached
        for payload in payloads:
            test_round.process_payload(payload)

        assert test_round.negative_vote_threshold_reached

    def test_positive_threshold(
        self,
    ) -> None:
        """Runs test."""

        test_round = self.setup_test_voting_round()
        first_payload, *payloads = get_dummy_tx_payloads(self.participants, vote=True)
        test_round.process_payload(first_payload)

        assert not test_round.positive_vote_threshold_reached
        for payload in payloads:
            test_round.process_payload(payload)

        assert test_round.positive_vote_threshold_reached


class TestCollectDifferentUntilThresholdRound(_BaseRoundTestClass):
    """Test CollectDifferentUntilThresholdRound."""

    @pytest.mark.parametrize(
        "required_confirmations", (MAX_PARTICIPANTS, MAX_PARTICIPANTS + 1)
    )
    def test_run(
        self,
        required_confirmations: int,
    ) -> None:
        """Run tests."""

        test_round = DummyCollectDifferentUntilThresholdRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )
        test_round.block_confirmations = 0
        test_round.required_block_confirmations = required_confirmations
        test_round.selection_key = "selection_key"
        test_round.collection_key = "collection_key"
        test_round.done_event = 0
        test_round.no_majority_event = 1
        assert (
            test_round.consensus_threshold <= required_confirmations
        ), "Incorrect test parametrization: required confirmations cannot be set with a smalled value than the consensus threshold"

        first_payload, *payloads = get_dummy_tx_payloads(self.participants, vote=False)
        test_round.process_payload(first_payload)

        assert not test_round.collection_threshold_reached
        for payload in payloads:
            test_round.process_payload(payload)
            res = test_round.end_block()
            if test_round.block_confirmations <= required_confirmations:
                assert res is None
            else:
                assert res is not None
                assert res[1] == test_round.done_event
        assert test_round.collection_threshold_reached
        self._test_payload_with_wrong_round_count(test_round)


class TestCollectNonEmptyUntilThresholdRound(_BaseRoundTestClass):
    """Test `CollectNonEmptyUntilThresholdRound`."""

    def test_get_non_empty_values(self) -> None:
        """Test `_get_non_empty_values`."""
        test_round = DummyCollectNonEmptyUntilThresholdRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )
        payloads = get_dummy_tx_payloads(self.participants)
        payloads[3]._value = None
        for payload in payloads:
            test_round.process_payload(payload)

        non_empty_values = test_round._get_non_empty_values()
        assert non_empty_values == [f"agent_{i}" for i in range(3)]

        self._test_payload_with_wrong_round_count(test_round)

    def test_process_payload(self) -> None:
        """Test `process_payload`."""
        test_round = DummyCollectNonEmptyUntilThresholdRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        first_payload, *payloads = get_dummy_tx_payloads(self.participants)
        test_round.process_payload(first_payload)

        assert not test_round.collection_threshold_reached
        for payload in payloads:
            test_round.process_payload(payload)

        assert test_round.collection_threshold_reached

    @pytest.mark.parametrize("is_majority_possible", (True, False))
    def test_end_block_no_threshold_reached(self, is_majority_possible: bool) -> None:
        """Test `end_block` when no collection threshold is reached."""
        test_round = DummyCollectNonEmptyUntilThresholdRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        test_round.is_majority_possible = lambda *_: is_majority_possible  # type: ignore
        test_round.no_majority_event = "no_majority"

        res = cast(Tuple[BaseSynchronizedData, Enum], test_round.end_block())

        if test_round.block_confirmations > test_round.required_block_confirmations:
            assert res[0].db == self.synchronized_data.db
            assert res[1] == test_round.no_majority_event
        else:
            assert res is None

    @pytest.mark.parametrize(
        "is_value_none, expected_event", ((True, "none"), (False, "done"))
    )
    def test_end_block(self, is_value_none: bool, expected_event: str) -> None:
        """Test `end_block` when collection threshold is reached."""
        test_round = DummyCollectNonEmptyUntilThresholdRound(
            synchronized_data=self.synchronized_data,
            consensus_params=self.consensus_params,
        )

        payloads = get_dummy_tx_payloads(self.participants, is_value_none=is_value_none)
        for payload in payloads:
            test_round.process_payload(payload)

        test_round.collection = {f"test_{i}": payloads[i] for i in range(len(payloads))}
        test_round.selection_key = "test"
        test_round.collection_key = "test"
        test_round.done_event = "done"
        test_round.none_event = "none"

        res = cast(Tuple[BaseSynchronizedData, Enum], test_round.end_block())
        assert res[0].db == self.synchronized_data.db
        assert res[1] == expected_event



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests package for the 'valory/contract_api' protocol."""
from abc import abstractmethod
from typing import Callable, Type
from unittest import mock

import pytest
from aea.common import Address
from aea.exceptions import AEAEnforceError
from aea.mail.base import Envelope
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import DialogueLabel

import packages
from packages.valory.protocols.contract_api import ContractApiMessage
from packages.valory.protocols.contract_api.custom_types import Kwargs
from packages.valory.protocols.contract_api.dialogues import (
    ContractApiDialogue,
    ContractApiDialogues,
)
from packages.valory.protocols.contract_api.message import (
    _default_logger as contract_api_message_logger,
)


LEDGER_ID = "ethereum"
CONTRACT_ID = "contract_id_stub"
CALLABLE = "callable_stub"
CONTRACT_ADDRESS = "contract_address_stub"


class BaseTestMessageConstruction:
    """Base class to test message construction for the ABCI protocol."""

    ledger_id = LEDGER_ID
    contract_id = CONTRACT_ID
    callable_ = CALLABLE
    msg_class = ContractApiMessage
    contract_address = CONTRACT_ADDRESS

    @abstractmethod
    def build_message(self) -> ContractApiMessage:
        """Build the message to be used for testing."""

    def test_run(self) -> None:
        """Run the test."""
        msg = self.build_message()
        msg.to = "receiver"
        envelope = Envelope(to=msg.to, sender="sender", message=msg)
        envelope_bytes = envelope.encode()

        actual_envelope = Envelope.decode(envelope_bytes)
        expected_envelope = envelope

        assert expected_envelope.to == actual_envelope.to
        assert expected_envelope.sender == actual_envelope.sender
        assert (
            expected_envelope.protocol_specification_id
            == actual_envelope.protocol_specification_id
        )
        assert expected_envelope.message != actual_envelope.message

        actual_msg = self.msg_class.serializer.decode(actual_envelope.message_bytes)
        actual_msg.to = actual_envelope.to
        actual_msg.sender = actual_envelope.sender
        expected_msg = msg
        assert expected_msg == actual_msg

    @classmethod
    def _make_kwargs(cls) -> Kwargs:
        """Build a ConsensuParams object."""
        return Kwargs(body={})


class TestGetDeployTransaction(BaseTestMessageConstruction):
    """Test message."""

    def build_message(self) -> ContractApiMessage:
        """Build the message."""
        assert str(self._make_kwargs()) == "Kwargs: body={}"
        return ContractApiMessage(
            performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,  # type: ignore
            ledger_id=self.ledger_id,
            contract_id=self.contract_id,
            callable=self.callable_,
            kwargs=self._make_kwargs(),
        )


class TestError(BaseTestMessageConstruction):
    """Test message."""

    def build_message(self) -> ContractApiMessage:
        """Build the message."""
        return ContractApiMessage(
            performative=ContractApiMessage.Performative.ERROR,  # type: ignore
            code=1,
            message="",
            data=b"",
        )


class TestGetRawTransaction(BaseTestMessageConstruction):
    """Test message."""

    def build_message(self) -> ContractApiMessage:
        """Build the message."""
        return ContractApiMessage(
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            ledger_id=self.ledger_id,
            contract_id=self.contract_id,
            contract_address=self.contract_address,
            callable=self.callable_,
            kwargs=self._make_kwargs(),
        )


class TestRawTransaction(BaseTestMessageConstruction):
    """Test message."""

    def build_message(self) -> ContractApiMessage:
        """Build the message."""
        return ContractApiMessage(
            performative=ContractApiMessage.Performative.RAW_TRANSACTION,  # type: ignore
            raw_transaction=ContractApiMessage.RawTransaction(
                ledger_id=LEDGER_ID, body={}
            ),
        )


class TestGetRawMessage(BaseTestMessageConstruction):
    """Test message."""

    def build_message(self) -> ContractApiMessage:
        """Build the message."""
        return ContractApiMessage(
            performative=ContractApiMessage.Performative.GET_RAW_MESSAGE,  # type: ignore
            ledger_id=self.ledger_id,
            contract_id=self.contract_id,
            contract_address=self.contract_address,
            callable=self.callable_,
            kwargs=self._make_kwargs(),
        )


class TestRawMessage(BaseTestMessageConstruction):
    """Test message."""

    def build_message(self) -> ContractApiMessage:
        """Build the message."""
        return ContractApiMessage(
            performative=ContractApiMessage.Performative.RAW_MESSAGE,  # type: ignore
            raw_message=ContractApiMessage.RawMessage(ledger_id=LEDGER_ID, body=b""),
        )


class TestGetState(BaseTestMessageConstruction):
    """Test message."""

    def build_message(self) -> ContractApiMessage:
        """Build the message."""
        return ContractApiMessage(
            performative=ContractApiMessage.Performative.GET_STATE,  # type: ignore
            ledger_id=self.ledger_id,
            contract_id=self.contract_id,
            contract_address=self.contract_address,
            callable=self.callable_,
            kwargs=self._make_kwargs(),
        )


class TestState(BaseTestMessageConstruction):
    """Test message."""

    def build_message(self) -> ContractApiMessage:
        """Build the message."""
        return ContractApiMessage(
            performative=ContractApiMessage.Performative.STATE,  # type: ignore
            state=ContractApiMessage.State(ledger_id=LEDGER_ID, body={}),
        )


@mock.patch.object(
    packages.valory.protocols.contract_api.message,
    "enforce",
    side_effect=AEAEnforceError("some error"),
)
def test_incorrect_message(mocked_enforce: Callable) -> None:
    """Test that we raise an exception when the message is incorrect."""
    with mock.patch.object(contract_api_message_logger, "error") as mock_logger:
        ContractApiMessage(
            message_id=1,
            dialogue_reference=(str(0), ""),
            target=0,
            performative=ContractApiMessage.Performative.RAW_MESSAGE,  # type: ignore
            raw_message=ContractApiMessage.RawMessage("some_ledger_id", b"some_body"),
        )

        mock_logger.assert_any_call("some error")


def test_performative_string_value() -> None:
    """Test the string valoe of performatives."""

    assert (
        str(ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION)
        == "get_deploy_transaction"
    ), "The str value must be get_deploy_transaction"


def test_encoding_unknown_performative() -> None:
    """Test that we raise an exception when the performative is unknown during encoding."""
    msg = ContractApiMessage(
        performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,  # type: ignore
        ledger_id=LEDGER_ID,
        contract_id=CONTRACT_ID,
        callable=CALLABLE,
        kwargs=Kwargs({}),
    )
    with pytest.raises(ValueError, match="Performative not valid:"):
        with mock.patch.object(
            ContractApiMessage.Performative, "__eq__", return_value=False
        ):
            ContractApiMessage.serializer.encode(msg)


def test_decoding_unknown_performative() -> None:
    """Test that we raise an exception when the performative is unknown during encoding."""
    msg = ContractApiMessage(
        performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,  # type: ignore
        ledger_id=LEDGER_ID,
        contract_id=CONTRACT_ID,
        callable=CALLABLE,
        kwargs=Kwargs({}),
    )

    encoded_msg = ContractApiMessage.serializer.encode(msg)
    with pytest.raises(ValueError, match="Performative not valid:"):
        with mock.patch.object(
            ContractApiMessage.Performative, "__eq__", return_value=False
        ):
            ContractApiMessage.serializer.decode(encoded_msg)


class AgentDialogue(ContractApiDialogue):
    """The dialogue class maintains state of a dialogue and manages it."""

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: BaseDialogue.Role,
        message_class: Type[ContractApiMessage],
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for
        :param message_class: the message class
        """
        ContractApiDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class AgentDialogues(ContractApiDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address) -> None:
        """
        Initialize dialogues.

        :param self_address: the address of the entity for whom this dialogue is maintained
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return ContractApiDialogue.Role.AGENT

        ContractApiDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=AgentDialogue,
        )


class LedgerDialogue(ContractApiDialogue):
    """The dialogue class maintains state of a dialogue and manages it."""

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: BaseDialogue.Role,
        message_class: Type[ContractApiMessage],
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for
        :param message_class: the message class
        """
        ContractApiDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class LedgerDialogues(ContractApiDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address) -> None:
        """
        Initialize dialogues.

        :param self_address: the address of the entity for whom this dialogue is maintained
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return ContractApiDialogue.Role.LEDGER

        ContractApiDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=LedgerDialogue,
        )


class TestDialogues:
    """Tests abci dialogues."""

    agent_addr: str
    ledger_addr: str
    agent_dialogues: AgentDialogues
    ledger_dialogues: LedgerDialogues

    @classmethod
    def setup_class(cls) -> None:
        """Set up the test."""
        cls.agent_addr = "agent address"
        cls.ledger_addr = "ledger address"
        cls.agent_dialogues = AgentDialogues(cls.agent_addr)
        cls.ledger_dialogues = LedgerDialogues(cls.ledger_addr)

    def test_create_self_initiated(self) -> None:
        """Test the self initialisation of a dialogue."""
        result = self.agent_dialogues._create_self_initiated(
            dialogue_opponent_addr=self.ledger_addr,
            dialogue_reference=(str(0), ""),
            role=ContractApiDialogue.Role.AGENT,
        )
        assert isinstance(result, ContractApiDialogue)
        assert result.role == ContractApiDialogue.Role.AGENT, "The role must be agent."

    def test_create_opponent_initiated(self) -> None:
        """Test the opponent initialisation of a dialogue."""
        result = self.agent_dialogues._create_opponent_initiated(
            dialogue_opponent_addr=self.ledger_addr,
            dialogue_reference=(str(0), ""),
            role=ContractApiDialogue.Role.AGENT,
        )
        assert isinstance(result, ContractApiDialogue)
        assert result.role == ContractApiDialogue.Role.AGENT, "The role must be agent."



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: http.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\nhttp.proto\x12\x16\x61\x65\x61.valory.http.v1_0_0"\x91\x03\n\x0bHttpMessage\x12K\n\x07request\x18\x05 \x01(\x0b\x32\x38.aea.valory.http.v1_0_0.HttpMessage.Request_PerformativeH\x00\x12M\n\x08response\x18\x06 \x01(\x0b\x32\x39.aea.valory.http.v1_0_0.HttpMessage.Response_PerformativeH\x00\x1a\x63\n\x14Request_Performative\x12\x0e\n\x06method\x18\x01 \x01(\t\x12\x0b\n\x03url\x18\x02 \x01(\t\x12\x0f\n\x07version\x18\x03 \x01(\t\x12\x0f\n\x07headers\x18\x04 \x01(\t\x12\x0c\n\x04\x62ody\x18\x05 \x01(\x0c\x1aq\n\x15Response_Performative\x12\x0f\n\x07version\x18\x01 \x01(\t\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x13\n\x0bstatus_text\x18\x03 \x01(\t\x12\x0f\n\x07headers\x18\x04 \x01(\t\x12\x0c\n\x04\x62ody\x18\x05 \x01(\x0c\x42\x0e\n\x0cperformativeb\x06proto3'
)


_HTTPMESSAGE = DESCRIPTOR.message_types_by_name["HttpMessage"]
_HTTPMESSAGE_REQUEST_PERFORMATIVE = _HTTPMESSAGE.nested_types_by_name[
    "Request_Performative"
]
_HTTPMESSAGE_RESPONSE_PERFORMATIVE = _HTTPMESSAGE.nested_types_by_name[
    "Response_Performative"
]
HttpMessage = _reflection.GeneratedProtocolMessageType(
    "HttpMessage",
    (_message.Message,),
    {
        "Request_Performative": _reflection.GeneratedProtocolMessageType(
            "Request_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _HTTPMESSAGE_REQUEST_PERFORMATIVE,
                "__module__": "http_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.http.v1_0_0.HttpMessage.Request_Performative)
            },
        ),
        "Response_Performative": _reflection.GeneratedProtocolMessageType(
            "Response_Performative",
            (_message.Message,),
            {
                "DESCRIPTOR": _HTTPMESSAGE_RESPONSE_PERFORMATIVE,
                "__module__": "http_pb2"
                # @@protoc_insertion_point(class_scope:aea.valory.http.v1_0_0.HttpMessage.Response_Performative)
            },
        ),
        "DESCRIPTOR": _HTTPMESSAGE,
        "__module__": "http_pb2"
        # @@protoc_insertion_point(class_scope:aea.valory.http.v1_0_0.HttpMessage)
    },
)
_sym_db.RegisterMessage(HttpMessage)
_sym_db.RegisterMessage(HttpMessage.Request_Performative)
_sym_db.RegisterMessage(HttpMessage.Response_Performative)

if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _HTTPMESSAGE._serialized_start = 39
    _HTTPMESSAGE._serialized_end = 440
    _HTTPMESSAGE_REQUEST_PERFORMATIVE._serialized_start = 210
    _HTTPMESSAGE_REQUEST_PERFORMATIVE._serialized_end = 309
    _HTTPMESSAGE_RESPONSE_PERFORMATIVE._serialized_start = 311
    _HTTPMESSAGE_RESPONSE_PERFORMATIVE._serialized_end = 424
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tendermint Docker image."""
import os
import re
import subprocess  # nosec
import time
from pathlib import Path
from typing import Any, Dict, List

import docker
import pytest
from docker.errors import ImageNotFound
from docker.models.containers import Container

from autonomy.constants import TENDERMINT_IMAGE_NAME, TENDERMINT_IMAGE_VERSION

from tests.helpers.base import tendermint_health_check
from tests.helpers.docker.base import DockerImage


_TCP = "tcp://"
_HTTP = "http://"
_LOCAL_ADDRESS = "0.0.0.0"  # nosec

DEFAULT_TENDERMINT_PORT = 26657
DEFAULT_P2P_PORT = 26656
DEFAULT_TENDERMINT_COM_PORT = 8080
DEFAULT_ABCI_PORT = 26658
# we need this because we want to connect from the Tendermint
# Docker container to the ABCI server that lives in the host
DEFAULT_ABCI_HOST = "host.docker.internal"

_SLEEP_TIME = 1


class TendermintDockerImage(DockerImage):
    """Tendermint Docker image."""

    def __init__(
        self,
        client: docker.DockerClient,
        abci_host: str = DEFAULT_ABCI_HOST,
        abci_port: int = DEFAULT_ABCI_PORT,
        port: int = DEFAULT_TENDERMINT_PORT,
        p2p_port: int = DEFAULT_P2P_PORT,
        com_port: int = DEFAULT_TENDERMINT_COM_PORT,
    ):
        """Initialize."""
        super().__init__(client)
        self.abci_host = abci_host
        self.abci_port = abci_port
        self.port = port
        self.p2p_port = p2p_port
        self.com_port = com_port
        self.proxy_app = f"{_TCP}{self.abci_host}:{self.abci_port}"

    @property
    def tag(self) -> str:
        """Get the tag."""
        return "tendermint/tendermint:v0.34.11"

    def _build_command(self) -> List[str]:
        """Build command."""
        cmd = ["node", f"--proxy_app={self.proxy_app}"]
        return cmd

    def create(self) -> Container:
        """Create the container."""
        cmd = self._build_command()
        ports = {
            f"{DEFAULT_TENDERMINT_PORT}/tcp": (_LOCAL_ADDRESS, self.port),  # nosec
        }
        if self.abci_host == DEFAULT_ABCI_HOST:
            extra_hosts_config = {self.abci_host: "host-gateway"}
        else:
            extra_hosts_config = {}
        container = self._client.containers.run(
            self.tag,
            command=cmd,
            detach=True,
            ports=ports,
            extra_hosts=extra_hosts_config,
        )
        return container

    def create_many(self, nb_containers: int) -> List[Container]:
        """Instantiate the image in many containers, parametrized."""
        raise NotImplementedError()

    def wait(self, max_attempts: int = 15, sleep_rate: float = 1.0) -> bool:
        """
        Wait until the image is running.

        :param max_attempts: max number of attempts.
        :param sleep_rate: the amount of time to sleep between different requests.
        :return: True if the wait was successful, False otherwise.
        """
        time.sleep(_SLEEP_TIME)
        return True


class FlaskTendermintDockerImage(TendermintDockerImage):
    """Flask app with Tendermint Docker image."""

    _extra_hosts: Dict[str, str]

    def __init__(
        self,
        client: docker.DockerClient,
        abci_host: str = DEFAULT_ABCI_HOST,
        abci_port: int = DEFAULT_ABCI_PORT,
        port: int = DEFAULT_TENDERMINT_PORT,
        p2p_port: int = DEFAULT_P2P_PORT,
        com_port: int = DEFAULT_TENDERMINT_COM_PORT,
    ):
        """Initialize."""
        super().__init__(client, abci_host, abci_port, port, p2p_port, com_port)
        self.__create_flask_tendermint_image()

    def __create_flask_tendermint_image(self) -> None:
        """Create an image of the Flask server with Tendermint."""
        try:
            self._client.images.get(self.tag)
        except ImageNotFound:
            cwd = os.getcwd()
            current_file_folder = os.path.dirname(os.path.realpath(__file__))
            root = current_file_folder.split(os.path.sep)[:-3]
            os.chdir(os.path.join(os.path.sep, *root))
            cmd = [
                "autonomy",
                "deploy",
                "build",
                "image",
                "valory/oracle_hardhat",
                "--dependencies",
            ]
            subprocess.run(cmd)  # nosec
            os.chdir(cwd)

    @property
    def tag(self) -> str:
        """Get the tag."""
        return f"{TENDERMINT_IMAGE_NAME}:{TENDERMINT_IMAGE_VERSION}"

    @staticmethod
    def get_node_name(i: int) -> str:
        """Get the ith node's name."""
        return f"node{i}"

    def _get_node_id(self, i: int) -> str:
        """Get the node id."""
        cmd = [
            "docker",
            "exec",
            self.get_node_name(i),
            "tendermint",
            "--home",
            self.get_node_name(i),
            "show-node-id",
        ]
        process = subprocess.Popen(  # nosec
            cmd,
            stdout=subprocess.PIPE,
        )
        output, _ = process.communicate()
        node_id = output.decode().strip()
        return node_id

    @staticmethod
    def __increment_port(port: int, i: int) -> int:
        """Increment a port"""
        return port + i * 10

    def get_port(self, i: int) -> int:
        """Get the ith port."""
        return self.__increment_port(self.port, i)

    def get_com_port(self, i: int) -> int:
        """Get the ith com port."""
        return self.__increment_port(self.com_port, i)

    def get_p2p_port(self, i: int) -> int:
        """Get the ith p2p port."""
        return self.__increment_port(self.p2p_port, i)

    def get_abci_port(self, i: int) -> int:
        """Get the ith abci port."""
        return self.__increment_port(self.abci_port, i)

    def get_addr(self, prefix: str, i: int, p2p: bool = False) -> str:
        """Get a node's address."""
        valid_prefixes = {_TCP, _HTTP}
        if prefix not in valid_prefixes:
            raise ValueError(f"Invalid prefix! Should be one of: {valid_prefixes}")

        if p2p:
            prefix += self._get_node_id(i) + "@"
            port = self.get_p2p_port(i)
        else:
            port = self.get_port(i)

        return f"{prefix}{_LOCAL_ADDRESS}:{port}"

    @property
    def p2p_seeds(self) -> List[str]:
        """Get p2p seeds."""
        if self.nb_nodes is None:
            raise ValueError("Trying to get p2p seeds before initializing containers!")

        return [self.get_addr(_TCP, i) for i in range(self.nb_nodes)]

    def _build_command(self) -> List[str]:
        """Build command."""
        return ["run", "--no-reload", f"--host={_LOCAL_ADDRESS}", "--port=8080"]

    def _create_one(self, i: int) -> Container:
        """Create a node container."""
        name = self.get_node_name(i)
        extra_hosts = (
            {self.abci_host: "host-gateway"}
            if self.abci_host == DEFAULT_ABCI_HOST
            else {}
        )
        extra_hosts.update(self._extra_hosts)

        run_kwargs = dict(
            image=self.tag,
            command=self._build_command(),
            name=name,
            hostname=name,
            detach=True,
            mem_limit="1024m",
            mem_reservation="256M",
            environment={
                "ID": i,
                "PROXY_APP": f"{_TCP}{self.abci_host}:{self.get_abci_port(i)}",
                "TMHOME": f"/tendermint/{name}",
                "CREATE_EMPTY_BLOCKS": "true",
                "DEV_MODE": "1",
                "LOG_FILE": f"/logs/{name}.txt",
            },
            working_dir="/tendermint",
            volumes=[
                f"{os.getcwd()}/nodes:/tendermint:Z",
                f"{os.getcwd()}/logs:/logs:Z",
                f"{os.getcwd()}/tm_state:/tm_state:Z",
            ],
            ports={
                f"{DEFAULT_TENDERMINT_PORT}/tcp": (
                    _LOCAL_ADDRESS,  # nosec,
                    self.get_port(i),
                ),
                f"{DEFAULT_TENDERMINT_COM_PORT}/tcp": (
                    _LOCAL_ADDRESS,  # nosec,
                    self.get_com_port(i),
                ),
                f"{DEFAULT_P2P_PORT}/tcp": (
                    _LOCAL_ADDRESS,  # nosec,
                    self.get_p2p_port(i),
                ),
            },
            extra_hosts=extra_hosts,
        )
        container = self._client.containers.run(**run_kwargs)
        return container

    def _create_config(self, nb_nodes: int) -> None:
        """Create necessary configuration."""
        self.nb_nodes = nb_nodes
        cmd = [
            "docker",
            "run",
            "--rm",
            "-v",
            f"{os.getcwd()}/nodes:/tendermint:Z",
            "--entrypoint=/usr/bin/tendermint",
            self.tag,
            "testnet",
            "--config",
            "/etc/tendermint/config-template.toml",
            "--v",
            f"{nb_nodes}",
            "--o",
            "/tendermint/",
        ]
        for i in range(nb_nodes):
            cmd.append(f"--hostname=node{i}")

        subprocess.run(cmd)  # nosec
        for config_file in Path().cwd().glob("nodes/**/*.toml"):
            config_text = config_file.read_text(encoding="utf-8")
            peers = re.findall(r"[a-z0-9]+@node\d:\d+", config_text)

            updated_peers = []
            for peer in peers:
                peer_id, address = peer.split("@")
                peer_name, port_string = address.split(":")
                *_, peer_number_string = peer_name

                port = int(port_string)
                peer_number = int(peer_number_string)
                new_port = port + (peer_number * 10)
                updated_peers.append(f"{peer_id}@{peer_name}:{new_port}")

            persistent_peers_string = (
                'persistent_peers = "' + ",".join(updated_peers) + '"\n'
            )
            updated_config = re.sub(
                'persistent_peers = ".*\n', persistent_peers_string, config_text
            )

            config_file.write_text(updated_config, encoding="utf-8")

        self._extra_hosts = {
            self.get_node_name(i): "host-gateway" for i in range(nb_nodes)
        }

    def create_many(self, nb_containers: int) -> List[Container]:
        """Create a list of node containers."""
        self._create_config(nb_containers)
        containers = [self._create_one(i) for i in range(nb_containers)]
        return containers

    def health_check(self, **kwargs: Any) -> None:
        """Do a health-check of the Tendermint network."""
        http_rpc_laddresses = [self.get_addr(_HTTP, i) for i in range(self.nb_nodes)]
        for http_rpc_laddr in http_rpc_laddresses:
            if not tendermint_health_check(http_rpc_laddr, **kwargs):
                pytest.fail(
                    f"Tendermint node {http_rpc_laddr} did not pass health-check"
                )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Utils module."""

import json
from pathlib import Path


def fix_address_books(build_dir: Path) -> None:
    """Update address books in data dump to use them in replays."""
    for addr_file in sorted(
        (build_dir / "persistent_data" / "dumps").glob("**/addrbook.json")
    ):
        addr_data = json.loads(addr_file.read_text())
        for i in range(len(addr_data["addrs"])):
            *_, post_fix = addr_data["addrs"][i]["addr"]["ip"].split(".")
            addr_data["addrs"][i]["addr"]["ip"] = "127.0.0.1"
            addr_data["addrs"][i]["addr"]["port"] = int(f"2663{int(post_fix)-3}")

        addr_file.write_text(json.dumps(addr_data, indent=4))
        print(f"Updated {addr_file}")


def fix_config_files(build_dir: Path) -> None:
    """Update config.toml in data dump to use them in replays."""
    for config_file in sorted(
        (build_dir / "persistent_data" / "dumps").glob("**/config.toml")
    ):
        config = config_file.read_text()
        config = config.replace("persistent_peers =", "# persistent_peers =")
        config_file.write_text(config)
        print(f"Updated {config_file}")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the handlers for the 'counter_client' skill."""
# isort: skip_file # noqa
from typing import Optional, cast

from aea.configurations.data_types import PublicId
from aea.protocols.base import Message
from aea.skills.base import Handler

from packages.valory.protocols.http import (  # type: ignore # pylint: disable=no-name-in-module,import-error
    HttpMessage,
)
from packages.valory.skills.counter_client.dialogues import HttpDialogue, HttpDialogues
from packages.valory.skills.counter_client.utils import curdatetime, decode_value


class HttpHandler(Handler):
    """The HTTP response handler."""

    SUPPORTED_PROTOCOL = HttpMessage.protocol_id  # type: Optional[PublicId]

    def setup(self) -> None:
        """Set up the handler."""

    def handle(self, message: Message) -> None:
        """Handle a message."""
        http_message = cast(HttpMessage, message)

        # recover dialogue
        http_dialogues = cast(HttpDialogues, self.context.http_dialogues)
        http_dialogue = cast(
            Optional[HttpDialogue], http_dialogues.update(http_message)
        )
        if http_dialogue is None:
            self.context.logger.error(
                "something went wrong when adding the incoming HTTP message to the dialogue."
            )
            return

        if (
            http_message.performative != HttpMessage.Performative.RESPONSE
            or http_message.status_code != 200
        ):
            self.context.logger.info(
                f"response not valid: performative={http_message.performative}, "
                f"status_code={http_message.status_code}"
            )
            return

        request_nonce = http_dialogue.dialogue_label.dialogue_reference[0]
        callback_name = self.context.state.request_to_handler.pop(request_nonce, None)
        if callback_name is None:  # pragma: nocover
            return
        callback = getattr(self, callback_name, None)
        if callback is None:  # pragma: nocover
            return
        callback(http_message)

    def abci_query(self, message: HttpMessage) -> None:
        """Handle abci_query responses."""
        value = decode_value(message)
        has_changed = value != self.context.state.count
        self.context.logger.info(
            f"[{curdatetime()}] The counter value is: {value}."
            + ("Updating current state" if has_changed else "")
        )
        if has_changed:
            self.context.state.count = value

    def broadcast_tx_commit(self, _message: HttpMessage) -> None:
        """Handle broadcast_tx_commit responses."""
        self.context.logger.debug(f"[{curdatetime()}] Transaction completed.")

    def teardown(self) -> None:
        """Tear down the handler."""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test the handlers.py module of the skill."""
from typing import Any, cast
from unittest import mock
from unittest.mock import MagicMock

import pytest
from aea.configurations.data_types import PublicId

from packages.valory.protocols.abci import AbciMessage
from packages.valory.protocols.abci.custom_types import (
    CheckTxType,
    CheckTxTypeEnum,
    ValidatorUpdates,
)
from packages.valory.protocols.http import HttpMessage
from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AddBlockError,
    ERROR_CODE,
    OK_CODE,
    SignatureNotValidError,
)
from packages.valory.skills.abstract_round_abci.dialogues import (
    AbciDialogue,
    AbciDialogues,
)
from packages.valory.skills.abstract_round_abci.handlers import (
    ABCIRoundHandler,
    AbstractResponseHandler,
    exception_to_info_msg,
)


def test_exception_to_info_msg() -> None:
    """Test 'exception_to_info_msg' helper function."""
    exception = Exception("exception message")
    expected_string = f"{exception.__class__.__name__}: {str(exception)}"
    actual_string = exception_to_info_msg(exception)
    assert expected_string == actual_string


class TestABCIRoundHandler:
    """Test 'ABCIRoundHandler'."""

    def setup(self) -> None:
        """Set up the tests."""
        self.context = MagicMock(skill_id=PublicId.from_str("dummy/skill:0.1.0"))
        self.dialogues = AbciDialogues(name="", skill_context=self.context)
        self.handler = ABCIRoundHandler(name="", skill_context=self.context)

    def test_info(self) -> None:
        """Test the 'info' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_INFO,
            version="",
            block_version=0,
            p2p_version=0,
        )
        response = self.handler.info(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_INFO

    @pytest.mark.parametrize("app_hash", (b"", b"test"))
    def test_init_chain(self, app_hash: bytes) -> None:
        """Test the 'init_chain' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_INIT_CHAIN,
            time=MagicMock,
            chain_id="test_chain_id",
            consensus_params=MagicMock(),
            validators=MagicMock(),
            app_state_bytes=b"",
            initial_height=10,
        )
        self.context.state.round_sequence.last_round_transition_root_hash = app_hash
        response = self.handler.init_chain(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_INIT_CHAIN
        assert response.validators == ValidatorUpdates([])
        assert response.app_hash == app_hash

    def test_begin_block(self) -> None:
        """Test the 'begin_block' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_BEGIN_BLOCK,
            hash=b"",
            header=MagicMock(),
            last_commit_info=MagicMock(),
            byzantine_validators=MagicMock(),
        )
        response = self.handler.begin_block(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_BEGIN_BLOCK

    @mock.patch("packages.valory.skills.abstract_round_abci.handlers.Transaction")
    def test_check_tx(self, *_: Any) -> None:
        """Test the 'check_tx' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_CHECK_TX,
            tx=b"",
            type=CheckTxType(CheckTxTypeEnum.NEW),
        )
        response = self.handler.check_tx(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_CHECK_TX
        assert response.code == OK_CODE

    @mock.patch(
        "packages.valory.skills.abstract_round_abci.handlers.Transaction.decode",
        side_effect=SignatureNotValidError,
    )
    def test_check_tx_negative(self, *_: Any) -> None:
        """Test the 'check_tx' handler method, negative case."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_CHECK_TX,
            tx=b"",
            type=CheckTxType(CheckTxTypeEnum.NEW),
        )
        response = self.handler.check_tx(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_CHECK_TX
        assert response.code == ERROR_CODE

    @mock.patch("packages.valory.skills.abstract_round_abci.handlers.Transaction")
    def test_deliver_tx(self, *_: Any) -> None:
        """Test the 'deliver_tx' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_DELIVER_TX,
            tx=b"",
        )
        response = self.handler.deliver_tx(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_DELIVER_TX
        assert response.code == OK_CODE

    @mock.patch(
        "packages.valory.skills.abstract_round_abci.handlers.Transaction.decode",
        side_effect=SignatureNotValidError,
    )
    def test_deliver_tx_negative(self, *_: Any) -> None:
        """Test the 'deliver_tx' handler method, negative case."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_DELIVER_TX,
            tx=b"",
        )
        response = self.handler.deliver_tx(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_DELIVER_TX
        assert response.code == ERROR_CODE

    @pytest.mark.parametrize("request_height", tuple(range(3)))
    def test_end_block(self, request_height: int) -> None:
        """Test the 'end_block' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_END_BLOCK,
            height=request_height,
        )
        assert isinstance(message, AbciMessage)
        assert isinstance(dialogue, AbciDialogue)
        assert message.height == request_height
        assert self.context.state.round_sequence.tm_height != request_height
        response = self.handler.end_block(message, dialogue)
        assert response.performative == AbciMessage.Performative.RESPONSE_END_BLOCK
        assert self.context.state.round_sequence.tm_height == request_height

    def test_commit(self) -> None:
        """Test the 'commit' handler method."""
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_COMMIT,
        )
        response = self.handler.commit(
            cast(AbciMessage, message), cast(AbciDialogue, dialogue)
        )
        assert response.performative == AbciMessage.Performative.RESPONSE_COMMIT

    def test_commit_negative(self) -> None:
        """Test the 'commit' handler method, negative case."""
        self.context.state.round_sequence.commit.side_effect = AddBlockError()
        message, dialogue = self.dialogues.create(
            counterparty="",
            performative=AbciMessage.Performative.REQUEST_COMMIT,
        )
        with pytest.raises(AddBlockError):
            self.handler.commit(
                cast(AbciMessage, message), cast(AbciDialogue, dialogue)
            )


class ConcreteResponseHandler(AbstractResponseHandler):
    """A concrete response handler for testing purposes."""

    SUPPORTED_PROTOCOL = HttpMessage.protocol_id
    allowed_response_performatives = frozenset({HttpMessage.Performative.RESPONSE})


class TestAbstractResponseHandler:
    """Test 'AbstractResponseHandler'."""

    def setup(self) -> None:
        """Set up the tests."""
        self.context = MagicMock()
        self.handler = ConcreteResponseHandler(name="", skill_context=self.context)

    def test_handle(self) -> None:
        """Test the 'handle' method."""
        callback = MagicMock()
        request_reference = "reference"
        self.context.requests.request_id_to_callback = {}
        self.context.requests.request_id_to_callback[request_reference] = callback
        with mock.patch.object(
            self.handler, "_recover_protocol_dialogues"
        ) as mock_dialogues_fn:
            mock_dialogue = MagicMock()
            mock_dialogue.dialogue_label.dialogue_reference = (request_reference, "")
            mock_dialogues = MagicMock()
            mock_dialogues.update = MagicMock(return_value=mock_dialogue)
            mock_dialogues_fn.return_value = mock_dialogues
            mock_message = MagicMock(performative=HttpMessage.Performative.RESPONSE)
            self.handler.handle(mock_message)
        callback.assert_called()

    @mock.patch.object(
        AbstractResponseHandler, "_recover_protocol_dialogues", return_value=None
    )
    def test_handle_negative_cannot_recover_dialogues(self, *_: Any) -> None:
        """Test the 'handle' method, negative case (cannot recover dialogues)."""
        self.handler.handle(MagicMock())

    @mock.patch.object(AbstractResponseHandler, "_recover_protocol_dialogues")
    def test_handle_negative_cannot_update_dialogues(
        self, mock_dialogues_fn: Any
    ) -> None:
        """Test the 'handle' method, negative case (cannot update dialogues)."""
        mock_dialogues = MagicMock(update=MagicMock(return_value=None))
        mock_dialogues_fn.return_value = mock_dialogues
        self.handler.handle(MagicMock())

    def test_handle_negative_performative_not_allowed(self) -> None:
        """Test the 'handle' method, negative case (performative not allowed)."""
        self.handler.handle(MagicMock())

    def test_handle_negative_cannot_find_callback(self) -> None:
        """Test the 'handle' method, negative case (cannot find callback)."""
        self.context.requests.request_id_to_callback = {}
        with pytest.raises(
            ABCIAppInternalError, match="No callback defined for request with nonce: "
        ):
            self.handler.handle(
                MagicMock(performative=HttpMessage.Performative.RESPONSE)
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test various price apis."""
import ast
import logging  # noqa: F401
from typing import Dict, List, Tuple

import requests

from packages.valory.skills.apy_estimation_abci.models import (
    FantomSubgraph,
    SpookySwapSubgraph,
)

from tests.test_skills.test_abstract_round_abci.test_models import DummyMessage
from tests.test_skills.test_apy_estimation_abci.conftest import (
    SpecsType,
    is_list_of_strings,
)


ResponseItemType = List[Dict[str, str]]
SubgraphResponseType = Dict[str, ResponseItemType]


def make_request(api_specs: Dict, query: str) -> requests.Response:
    """Make a request to a subgraph.

    :param api_specs: the subgraph's api_specs.
    :param query: the query.
    :return: a response dictionary.
    """
    if api_specs["method"] == "POST":
        r = requests.post(
            url=api_specs["url"], headers=api_specs["headers"], json={"query": query}
        )

        if r.status_code == 200:
            res = r.json()

            if (
                "errors" in res.keys()
                and res["errors"][0].get("locations", None) is not None
            ):
                message = res["errors"][0]["message"]
                location = res["errors"][0]["locations"][0]
                line = location["line"]
                column = location["column"]

                raise ValueError(
                    f"The given query is not correct.\nError in line {line}, column {column}: {message}"
                )

            elif "errors" in res.keys() or "data" not in res.keys():
                raise ValueError(f"Unknown error encountered!\nRaw response: {res}")

        else:
            raise ConnectionError(
                "Something went wrong while trying to communicate with the subgraph "
                f"(Error: {r.status_code})!\n{r.text}"
            )

    else:
        raise ValueError(
            f"Unknown method {api_specs['method']} for {api_specs['api_id']}"
        )

    return r


class TestSubgraphs:
    """Test for subgraphs."""

    @staticmethod
    def test_eth_price(spooky_specs: SpecsType, eth_price_usd_q: str) -> None:
        """Test SpookySwap's eth price request from subgraph."""
        spooky_specs["response_key"] += ":bundles"  # type: ignore
        api = SpookySwapSubgraph(**spooky_specs)

        res = make_request(api.get_spec(), eth_price_usd_q)
        eth_price = ast.literal_eval(api.process_response(DummyMessage(res.content))[0]["ethPrice"])  # type: ignore

        assert isinstance(eth_price, float)

    @staticmethod
    def test_block_from_timestamp(
        fantom_specs: SpecsType, block_from_timestamp_q: str
    ) -> None:
        """Test Fantom's block from timestamp request from subgraph."""
        fantom_specs["response_key"] += ":blocks"  # type: ignore
        api = FantomSubgraph(**fantom_specs)

        res = make_request(api.get_spec(), block_from_timestamp_q)
        block = api.process_response(DummyMessage(res.content))[0]  # type: ignore

        assert isinstance(block, dict)
        keys = ["number", "timestamp"]
        assert all((key in block for key in keys))
        assert all(isinstance(ast.literal_eval(block[key]), int) for key in keys)

    @staticmethod
    def test_top_n_pairs(spooky_specs: SpecsType, top_n_pairs_q: str) -> None:
        """Test SpookySwap's top n pairs request from subgraph."""
        spooky_specs["response_key"] += ":pairs"  # type: ignore
        api = SpookySwapSubgraph(**spooky_specs)
        api_specs = api.get_spec()
        api_specs["top_n_pools"] = 100

        res = make_request(api_specs, top_n_pairs_q)
        pair_ids = [pair["id"] for pair in api.process_response(DummyMessage(res.content))]  # type: ignore

        assert is_list_of_strings(pair_ids)

    @staticmethod
    def test_pairs(
        spooky_specs: SpecsType, pairs_q: str, pool_fields: Tuple[str, ...]
    ) -> None:
        """Test SpookySwap's pairs request from subgraph."""
        spooky_specs["response_key"] += ":pairs"  # type: ignore
        api = SpookySwapSubgraph(**spooky_specs)

        res = make_request(api.get_spec(), pairs_q)
        pairs = api.process_response(DummyMessage(res.content))  # type: ignore

        assert isinstance(pairs, list)
        assert len(pairs) > 0
        assert isinstance(pairs[0], dict)

        for pair in pairs:
            assert all((key in pair for key in pool_fields))



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the handler for the 'abci' skill."""
import struct
from typing import Any, cast

from packages.valory.protocols.abci import AbciMessage
from packages.valory.protocols.abci.custom_types import Events, ProofOps
from packages.valory.skills.abstract_abci.handlers import ABCIHandler
from packages.valory.skills.counter.dialogues import AbciDialogue


OK_CODE = 0
ERROR_CODE = 1


def encode_number(value: int) -> bytes:
    """Encode an integer (big-endian)."""
    return struct.pack(">I", value)


def decode_number(raw: bytes) -> int:
    """Decode an integer (little-endian)."""
    return int.from_bytes(raw, byteorder="big")


class ABCICounterHandler(ABCIHandler):
    """ABCI counter handler.

    Handles ABCI messages from a Tendermint node and implements the ABCI
    Counter app.
    """

    SUPPORTED_PROTOCOL = AbciMessage.protocol_id

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the handler."""
        super().__init__(*args, **kwargs)
        self.tx_count = 0
        self.last_block_height = 0

    def info(self, message: AbciMessage, dialogue: AbciDialogue) -> AbciMessage:
        """
        Handle a message of REQUEST_INFO performative.

        :param message: the ABCI request.
        :param dialogue: the ABCI dialogue.
        :return: the response.
        """
        reply = dialogue.reply(
            performative=AbciMessage.Performative.RESPONSE_INFO,
            target_message=message,
            info_data="",
            version=message.version,
            app_version=0,
            last_block_height=0,
            last_block_app_hash=b"",
        )
        return cast(AbciMessage, reply)

    def init_chain(self, message: AbciMessage, dialogue: AbciDialogue) -> AbciMessage:
        """
        Handle a message of REQUEST_INIT_CHAIN performative.

        :param message: the ABCI request.
        :param dialogue: the ABCI dialogue.
        :return: the response.
        """
        self.tx_count = 0
        self.last_block_height = 0
        return super().init_chain(message, dialogue)

    def query(self, message: AbciMessage, dialogue: AbciDialogue) -> AbciMessage:
        """
        Handle a message of REQUEST_QUERY performative.

        :param message: the ABCI request.
        :param dialogue: the ABCI dialogue.
        :return: the response.
        """
        value = encode_number(self.tx_count)
        log = f"value: {self.tx_count}"
        reply = dialogue.reply(
            performative=AbciMessage.Performative.RESPONSE_QUERY,
            code=OK_CODE,
            log=log,
            info="",
            index=0,
            key=b"",
            value=value,
            proof_ops=ProofOps([]),
            height=self.last_block_height,
            codespace="",
        )
        return cast(AbciMessage, reply)

    def check_tx(self, message: AbciMessage, dialogue: AbciDialogue) -> AbciMessage:
        """
        Handle a message of REQUEST_CHECK_TX performative.

        :param message: the ABCI request.
        :param dialogue: the ABCI dialogue.
        :return: the response.
        """
        value = decode_number(message.tx)
        if value == (self.tx_count + 1):
            code = OK_CODE
            log = "valid transaction."
            info = "OK: the next count is a unitary increment."
        else:
            code = ERROR_CODE
            log = "invalid transaction."
            info = "ERROR: the next count must be a unitary increment."

        reply = dialogue.reply(
            performative=AbciMessage.Performative.RESPONSE_CHECK_TX,
            code=code,
            data=b"",
            log=log,
            info=info,
            gas_wanted=0,
            gas_used=0,
            events=Events([]),
            codespace="",
        )
        return cast(AbciMessage, reply)

    def deliver_tx(self, message: AbciMessage, dialogue: AbciDialogue) -> AbciMessage:
        """
        Handle a message of REQUEST_DELIVER_TX performative.

        :param message: the ABCI request.
        :param dialogue: the ABCI dialogue.
        :return: the response.
        """
        self.tx_count += 1
        self.context.logger.info(f"The new count is: {self.tx_count}")
        reply = dialogue.reply(
            performative=AbciMessage.Performative.RESPONSE_DELIVER_TX,
            code=OK_CODE,
            data=b"",
            log="",
            info="",
            gas_wanted=0,
            gas_used=0,
            events=Events([]),
            codespace="",
        )
        return cast(AbciMessage, reply)

    def commit(self, message: AbciMessage, dialogue: AbciDialogue) -> AbciMessage:
        """
        Handle a message of REQUEST_COMMIT performative.

        :param message: the ABCI request.
        :param dialogue: the ABCI dialogue.
        :return: the response.
        """
        hash_ = struct.pack(">Q", self.tx_count)
        reply = dialogue.reply(
            performative=AbciMessage.Performative.RESPONSE_COMMIT,
            data=hash_,
            retain_height=0,
        )
        return cast(AbciMessage, reply)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Defines the base of the fuzzy tests"""
from typing import Dict, List, Tuple, Type

import numpy as np
from aea.exceptions import enforce
from aea.test_tools.test_cases import AEATestCaseMany
from hypothesis import given
from hypothesis.strategies import binary, booleans, integers, lists, text, tuples

from tests.test_connections.test_fuzz.mock_node.channels.base import BaseChannel
from tests.test_connections.test_fuzz.mock_node.node import MockNode


class BaseFuzzyTests(AEATestCaseMany):
    """
    Base class for the Fuzzy Tests
    """

    UINT_64_MAX_VALUE = np.iinfo(np.uint64).max
    UINT_64_MIN_VALUE = 0
    INT_64_MAX_VALUE = np.iinfo(np.int64).max
    INT_64_MIN_VALUE = np.iinfo(np.int64).min

    UINT_32_MAX_VALUE = np.iinfo(np.uint32).max
    UINT_32_MIN_VALUE = 0
    INT_32_MAX_VALUE = np.iinfo(np.int32).max
    INT_32_MIN_VALUE = np.iinfo(np.int32).min

    CHANNEL_TYPE: Type[BaseChannel] = BaseChannel
    CHANNEL_ARGS: Dict[str, any] = dict()  # type: ignore
    IS_LOCAL = True

    channel: BaseChannel
    mock_node: MockNode
    agent_package = "valory/test_abci:0.1.0"
    agent_name = "test_abci"
    agent_process = None
    cli_log_options = ["-v", "INFO"]

    AGENT_TIMEOUT = 10

    @classmethod
    def setUpClass(cls) -> None:
        """Sets up the environment for the tests."""
        cls.fetch_agent(cls.agent_package, cls.agent_name, is_local=cls.IS_LOCAL)
        cls.set_agent_context(cls.agent_name)
        cls.generate_private_key("ethereum", "ethereum_private_key.txt")
        cls.add_private_key("ethereum", "ethereum_private_key.txt")

        # we are mocking a tendermint node
        cls.set_config("vendor.valory.connections.abci.config.use_tendermint", False)

        cls.run_install()
        cls.agent_process = cls.run_agent()

        enforce(
            cls.is_running(cls.agent_process, cls.AGENT_TIMEOUT),
            f"The agent was not started in the defined timeout ({cls.AGENT_TIMEOUT}s)",
        )

        enforce(cls.CHANNEL_TYPE is not None, "A channel type must be provided")

        cls.channel = cls.CHANNEL_TYPE(**cls.CHANNEL_ARGS)
        cls.mock_node = MockNode(cls.channel)
        cls.mock_node.connect()

    @classmethod
    def tearDownClass(cls) -> None:
        """Tear down the testing environment."""
        cls.mock_node.disconnect()

    # flake8: noqa:D102
    @given(message=text())
    def test_echo(self, message: str) -> None:
        assert self.mock_node.echo(message)

    @given(
        version=text().filter(lambda x: x != ""),
        block_version=integers(0, UINT_64_MAX_VALUE),
        p2p_version=integers(0, UINT_64_MAX_VALUE),
    )
    def test_info(self, version: str, block_version: int, p2p_version: int) -> None:
        assert self.mock_node.info(version, block_version, p2p_version)

    def test_flush(self) -> None:
        assert self.mock_node.flush()

    @given(key=text(), value=text())
    def test_set_option(self, key: str, value: str) -> None:
        assert self.mock_node.set_option(key, value)

    @given(tx=binary())
    def test_deliver_tx(self, tx: bytes) -> None:
        assert self.mock_node.deliver_tx(tx)

    @given(tx=binary(), is_new_check=booleans())
    def test_check_tx(self, tx: bytes, is_new_check: bool) -> None:
        assert self.mock_node.check_tx(tx, is_new_check)

    @given(
        data=binary(),
        path=text(),
        height=integers(-INT_64_MAX_VALUE, INT_64_MAX_VALUE),
        prove=booleans(),
    )
    def test_query(self, data: bytes, path: str, height: int, prove: bool) -> None:
        assert self.mock_node.query(data, path, height, prove)

    def test_commit(self) -> None:
        assert self.mock_node.commit()

    @given(
        time_seconds=integers(0, INT_64_MAX_VALUE),
        time_nanos=integers(0, 10 ** 9),
        chain_id=text(),
        block_max_bytes=integers(1, INT_64_MAX_VALUE),
        block_max_gas=integers(-1, INT_32_MAX_VALUE),
        evidence_max_age_num_blocks=integers(INT_64_MIN_VALUE, INT_64_MAX_VALUE),
        evidence_max_age_seconds=integers(0, INT_64_MAX_VALUE),
        evidence_max_age_nanos=integers(0, 10 ** 9),
        evidence_max_bytes=integers(1, INT_64_MAX_VALUE),
        pub_key_types=lists(booleans()),
        app_version=integers(0, INT_32_MAX_VALUE),
        validator_pub_keys=lists(tuples(binary(), booleans())),
        validator_power=lists(integers(0, INT_32_MAX_VALUE)),
        app_state_bytes=binary(),
        initial_height=integers(0, INT_32_MAX_VALUE),
    )
    def test_init_chain(
        self,
        time_seconds: int,
        time_nanos: int,
        chain_id: str,
        block_max_bytes: int,
        block_max_gas: int,
        evidence_max_age_num_blocks: int,
        evidence_max_age_seconds: int,
        evidence_max_age_nanos: int,
        evidence_max_bytes: int,
        pub_key_types: List[str],
        app_version: int,
        validator_pub_keys: List[Tuple[bytes, bool]],
        validator_power: List[int],
        app_state_bytes: bytes,
        initial_height: int,
    ) -> None:
        min_validator_length = min(
            validator_power.__len__(),
            validator_pub_keys.__len__(),
            pub_key_types.__len__(),
        )

        assert self.mock_node.init_chain(
            time_seconds,
            time_nanos,
            chain_id,
            block_max_bytes,
            block_max_gas,
            evidence_max_age_num_blocks,
            evidence_max_age_seconds,
            evidence_max_age_nanos,
            evidence_max_bytes,
            ["ed25519" if pk else "secp256k1" for pk in pub_key_types][
                :min_validator_length
            ],
            app_version,
            [
                (pk, "ed25519") if tp else (pk, "secp256k1")
                for pk, tp in validator_pub_keys
            ][:min_validator_length],
            validator_power[:min_validator_length],
            app_state_bytes,
            initial_height,
        )

    @given(
        hash_=binary(),
        consen_ver_block=integers(0, UINT_64_MAX_VALUE),
        consen_ver_app=integers(0, UINT_64_MAX_VALUE),
        chain_id=text(),
        height=integers(-INT_64_MAX_VALUE, INT_64_MAX_VALUE),
        time_seconds=integers(1, INT_64_MAX_VALUE),
        time_nanos=integers(0, 10 ** 9),
        last_block_id_hash=binary(),
        last_commit_hash=binary(),
        data_hash=binary(),
        validators_hash=binary(),
        next_validators_hash=binary(),
        next_validators_part_header_total=integers(0, UINT_32_MAX_VALUE),
        next_validators_part_header_hash=binary(),
        header_consensus_hash=binary(),
        header_app_hash=binary(),
        header_last_results_hash=binary(),
        header_evidence_hash=binary(),
        header_proposer_address=binary(),
        last_commit_round=integers(INT_32_MIN_VALUE, INT_32_MAX_VALUE),
        last_commit_info_votes=lists(
            tuples(binary(), integers(INT_64_MIN_VALUE, INT_64_MAX_VALUE))
        ),
        last_commit_info_signed_last_block=lists(booleans()),
        evidence_type=lists(integers()),
        evidence_validator_address=lists(binary()),
        evidence_validator_power=lists(integers(INT_64_MIN_VALUE, INT_64_MAX_VALUE)),
        evidence_height=lists(integers(INT_64_MIN_VALUE, INT_64_MAX_VALUE)),
        evidence_time_seconds=lists(integers(1, INT_64_MAX_VALUE)),
        evidence_time_nanos=lists(integers(0, 10 ** 9)),
        evidence_total_voting_power=lists(integers(0, INT_64_MAX_VALUE)),
    )
    def test_begin_block(
        self,
        hash_: bytes,
        consen_ver_block: int,
        consen_ver_app: int,
        chain_id: str,
        height: int,
        time_seconds: int,
        time_nanos: int,
        last_block_id_hash: bytes,
        last_commit_hash: bytes,
        data_hash: bytes,
        validators_hash: bytes,
        next_validators_hash: bytes,
        next_validators_part_header_total: int,
        next_validators_part_header_hash: bytes,
        header_consensus_hash: bytes,
        header_app_hash: bytes,
        header_last_results_hash: bytes,
        header_evidence_hash: bytes,
        header_proposer_address: bytes,
        last_commit_round: int,
        last_commit_info_votes: List[Tuple[bytes, int]],
        last_commit_info_signed_last_block: List[bool],
        evidence_type: List[int],
        evidence_validator_address: List[bytes],
        evidence_validator_power: List[int],
        evidence_height: List[int],
        evidence_time_seconds: List[int],
        evidence_time_nanos: List[int],
        evidence_total_voting_power: List[int],
    ) -> None:
        last_commit_len = min(
            last_commit_info_votes.__len__(),
            last_commit_info_signed_last_block.__len__(),
        )
        evidence_len = min(
            evidence_type.__len__(),
            evidence_validator_address.__len__(),
            evidence_validator_power.__len__(),
            evidence_height.__len__(),
            evidence_time_seconds.__len__(),
            evidence_time_nanos.__len__(),
            evidence_total_voting_power.__len__(),
        )
        self.mock_node.begin_block(
            hash_,
            consen_ver_block,
            consen_ver_app,
            chain_id,
            height,
            time_seconds,
            time_nanos,
            last_block_id_hash,
            last_commit_hash,
            data_hash,
            validators_hash,
            next_validators_hash,
            next_validators_part_header_total,
            next_validators_part_header_hash,
            header_consensus_hash,
            header_app_hash,
            header_last_results_hash,
            header_evidence_hash,
            header_proposer_address,
            last_commit_round,
            last_commit_info_votes[:last_commit_len],
            last_commit_info_signed_last_block[:last_commit_len],
            evidence_type[:evidence_len],
            evidence_validator_address[:evidence_len],
            evidence_validator_power[:evidence_len],
            evidence_height[:evidence_len],
            evidence_time_seconds[:evidence_len],
            evidence_time_nanos[:evidence_len],
            evidence_total_voting_power[:evidence_len],
        )

    @given(height=integers(INT_64_MIN_VALUE, INT_64_MAX_VALUE))
    def test_end_block(self, height: int) -> None:
        assert self.mock_node.end_block(height)

    def test_list_snapshots(self) -> None:
        assert self.mock_node.list_snapshots()

    @given(
        height=integers(UINT_64_MIN_VALUE, UINT_64_MAX_VALUE),
        format_=integers(UINT_32_MIN_VALUE, UINT_32_MAX_VALUE),
        chunks=integers(UINT_32_MIN_VALUE, UINT_32_MAX_VALUE),
        hash_=binary(),
        metadata=binary(),
        app_hash=binary(),
    )
    def test_offer_snapshot(
        self,
        height: int,
        format_: int,
        chunks: int,
        hash_: bytes,
        metadata: bytes,
        app_hash: bytes,
    ) -> None:
        assert self.mock_node.offer_snapshot(
            height, format_, chunks, hash_, metadata, app_hash
        )

    @given(
        height=integers(UINT_64_MIN_VALUE, UINT_64_MAX_VALUE),
        format_=integers(UINT_32_MIN_VALUE, UINT_32_MAX_VALUE),
        chunk=integers(UINT_32_MIN_VALUE, UINT_32_MAX_VALUE),
    )
    def test_load_snapshot_chunk(self, height: int, format_: int, chunk: int) -> None:
        assert self.mock_node.load_snapshot_chunk(height, format_, chunk)

    @given(
        index=integers(UINT_32_MIN_VALUE, UINT_32_MAX_VALUE),
        chunk=binary(),
        sender=text(),
    )
    def test_apply_snapshot_chunk(self, index: int, chunk: bytes, sender: str) -> None:
        assert self.mock_node.apply_snapshot_chunk(index, chunk, sender)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------


"""Module for the queries building functions."""
import json
from typing import List, Optional


def finalize_q(query: str) -> bytes:
    """Finalize the given query string, i.e., add it under a `queries` key and convert it to bytes."""
    finalized_query = {"query": query}
    encoded_query = json.dumps(finalized_query, sort_keys=True).encode("utf-8")

    return encoded_query


def eth_price_usd_q(bundle_id: int, block: Optional[int] = None) -> bytes:
    """Create query to get current ETH price in USD.

    If `block` is given, then the query is being build for the corresponding block's price.

    :param bundle_id: the bundle id.
    :param block: the block for which the price will be fetched.
    :return: the built query.
    """

    if block is not None:
        query = (
            """
        {
            bundles(
                first: 1,
                block: {number: """
            + str(block)
            + """},
                where: {
                    id: """
            + str(bundle_id)
            + """
                }
            )
            {ethPrice}
        }
        """
        )

    else:
        query = (
            """
        {
            bundles(
                first: 1,
                where: {
                    id: """
            + str(bundle_id)
            + """
                }
            )
            {ethPrice}
        }
        """
        )

    return finalize_q(query)


def block_from_timestamp_q(timestamp: int) -> bytes:
    """Create query to get a block from a timestamp.

    :param timestamp: the timestamp for which the blocks will be fetched.
    :return: the built query.
    """
    query = (
        """
    {
        blocks(
            first: 1,
            orderBy: timestamp,
            orderDirection: asc,
            where: {
                timestamp_gte: """
        + str(timestamp)
        + """,
                timestamp_lte: """
        + str(timestamp + 600)
        + """
            }
        )
        {
            timestamp
            number
        }
    }
    """
    )

    return finalize_q(query)


def latest_block() -> bytes:
    """Create query to get the latest block.

    :return: the built query.
    """
    query = """
    {
        blocks(
            first: 1,
            orderBy: timestamp,
            orderDirection: desc
        )
        {
            timestamp
            number
        }
    }
    """

    return finalize_q(query)


def top_n_pairs_q(top_n: int) -> bytes:
    """Create a query to get the first `top_n` pool ids based on their total liquidity.

    :param top_n: the top_n pools to be fetched based on their total liquidity.
    :return: the built query.
    """

    query = (
        """
    {
        pairs(
            first: """
        + str(top_n)
        + """,
            orderBy: trackedReserveETH,
            orderDirection: desc
        )
        {id}
    }
    """
    )

    return finalize_q(query)


def pairs_q(block: int, ids: List[str]) -> bytes:
    """Create a query to get data for the given pools based on their total liquidity.

    :param block: the block for which the data are going to be fetched.
    :param ids: the ids of the pools to be fetched.
    :return: the built query.
    """

    query = (
        """
    {
        pairs(
            where: {id_in:
            [\""""
        + '","'.join(ids)
        + """"]},
            block: {number: """
        + str(block)
        + """}
        ) {
            id
            token0 {
                id
                symbol
                name
            }
            token1 {
                id
                symbol
                name
            }
            reserve0
            reserve1
            totalSupply
            reserveETH
            reserveUSD
            trackedReserveETH
            token0Price
            token1Price
            volumeToken0
            volumeToken1
            volumeUSD
            untrackedVolumeUSD
            txCount
            createdAtTimestamp
            createdAtBlockNumber
            liquidityProviderCount
        }
    }
    """
    )

    return finalize_q(query)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Config validators."""

import json

import jsonschema
from aea.configurations import validation
from aea.helpers.io import open_file

from autonomy.configurations.constants import SCHEMAS_DIR


class ConfigValidator(validation.ConfigValidator):
    """Configuration validator implementation."""

    def __init__(  # pylint: disable=super-init-not-called
        self, schema_filename: str, env_vars_friendly: bool = False
    ) -> None:
        """
        Initialize the parser for configuration files.

        :param schema_filename: the path to the JSON-schema file in 'aea/configurations/schemas'.
        :param env_vars_friendly: whether or not it is env var friendly.
        """

        with open_file(SCHEMAS_DIR / schema_filename) as fp:
            self._schema = json.load(fp)

        root_path = validation.make_jsonschema_base_uri(SCHEMAS_DIR)
        self._resolver = jsonschema.RefResolver(root_path, self._schema)
        self.env_vars_friendly = env_vars_friendly

        if env_vars_friendly:
            self._validator = validation.EnvVarsFriendlyDraft4Validator(
                self._schema, resolver=self._resolver
            )
        else:
            self._validator = validation.OwnDraft4Validator(
                self._schema, resolver=self._resolver
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for ETL."""
from typing import Any, Dict, List, Optional, Union

import pandas as pd
import pytest
from _pytest.monkeypatch import MonkeyPatch

from packages.valory.skills.apy_estimation_abci.tools.etl import (
    HIST_DTYPES,
    ResponseItemType,
    apply_revert_token_cols_wrapper,
    calc_apy,
    calc_change,
    prepare_batch,
    revert_transform_hist_data,
    transform_hist_data,
)
from packages.valory.skills.apy_estimation_abci.tools.io import (
    TRANSFORMED_HIST_DTYPES,
    load_hist,
)


class TestProcessing:
    """Tests for `Preprocessing`."""

    @staticmethod
    def test_calc_change() -> None:
        """Test `calc_change`."""
        test_series = pd.Series([1, 2, 3, 4, 5])
        expected_li: List[Optional[int]] = [1] * 4
        expected_li.insert(0, None)
        expected = pd.Series(expected_li)
        actual = calc_change(test_series)

        pd.testing.assert_series_equal(actual, expected)

    @staticmethod
    @pytest.mark.parametrize(
        "updated_reserve_usd,current_change,unrelated,expected",
        (
            (None, 1, 5.462345, None),
            (5, 1, "test", 14.6),
            (2, 10.3456, b"test", 377.6144),
            (7.34556, 1.45, 845, 14.410065400051186),
        ),
    )
    def test_calc_apy(
        updated_reserve_usd: Optional[float],
        current_change: float,
        unrelated: Any,
        expected: float,
    ) -> None:
        """Test `calc_apy`."""
        test_series = pd.Series(
            [updated_reserve_usd, current_change, unrelated],
            ["updatedReserveUSD", "currentChange", "unrelated"],
        )
        actual = calc_apy(test_series)

        if expected is None:
            assert pd.isna(actual)
        else:
            assert actual == expected

    @staticmethod
    def test_transform_hist_data(transformed_history: pd.DataFrame) -> None:
        """Test `transform_hist_data`."""
        # test with missing columns
        with pytest.raises(
            KeyError,
            match="Only a column name can be used for the key in a dtype mappings argument.",
        ):
            transform_hist_data([])

        # test with invalid types
        invalid_types: Dict[str, Union[int, str, Dict[str, str]]] = {
            key: 10 for key in HIST_DTYPES.keys()
        }
        with pytest.raises(TypeError):
            transform_hist_data([invalid_types])  # type: ignore

        # test with only one row, which leads to invalid `current_change`.
        invalid_current_change = invalid_types.copy()
        for i in range(2):
            invalid_current_change[f"token{i}"] = {
                "id": "x",
                "name": "x",
                "symbol": "X",
            }
        with pytest.raises(
            ValueError,
            match="APY cannot be calculated if there are not at least two observations for a pool!",
        ):
            transform_hist_data([invalid_current_change])  # type: ignore

        # test with normal input.
        data = invalid_current_change.copy()
        data["volumeUSD"] = 10
        test_pair_hist_raw = [data] * 2
        actual = transform_hist_data(test_pair_hist_raw)  # type: ignore
        pd.testing.assert_frame_equal(
            actual, transformed_history.astype(TRANSFORMED_HIST_DTYPES)
        )

    @staticmethod
    def test_load_hist(
        monkeypatch: MonkeyPatch, transformed_history: pd.DataFrame
    ) -> None:
        """Test `load_hist`."""
        monkeypatch.setattr(pd, "read_csv", lambda _: transformed_history)
        actual = load_hist("")

        actual_timestamp = actual.loc[1, "blockTimestamp"]
        expected_timestamp = pd.Timestamp(
            transformed_history.loc[1, "blockTimestamp"], unit="s"
        )
        assert expected_timestamp == actual_timestamp

        transformed_history = transformed_history.astype(TRANSFORMED_HIST_DTYPES)
        for df in (actual, transformed_history):
            df.drop(columns="blockTimestamp", inplace=True)
        pd.testing.assert_frame_equal(actual, transformed_history)

    @staticmethod
    def test_load_hist_file_not_found(
        monkeypatch: MonkeyPatch, transformed_history: pd.DataFrame
    ) -> None:
        """Test `load_hist` when file is not found."""
        with pytest.raises(IOError):
            load_hist("non_existing")

    @staticmethod
    @pytest.mark.parametrize("invalid", (True, False))
    def test_prepare_batch(
        monkeypatch: MonkeyPatch,
        transformed_historical_data_no_datetime_conversion: pd.DataFrame,
        batch: ResponseItemType,
        invalid: bool,
    ) -> None:
        """Test `prepare_batch`."""
        previous_batch = transformed_historical_data_no_datetime_conversion.iloc[
            [0, 2]
        ].reset_index(drop=True)

        if invalid:
            invalid_pool_item = batch[0].copy()
            non_existing_id = "non_existing"
            invalid_pool_item["id"] = non_existing_id
            batch.append(invalid_pool_item)
            with pytest.raises(
                ValueError, match="Could not find any previous history in "
            ):
                prepare_batch(previous_batch, batch)

        else:
            prepared_batches = prepare_batch(previous_batch, batch).groupby("id")
            assert len(prepared_batches) == 2
            expected_ids = {"0x2b4c76d0dc16be1c31d4c1dc53bf9b45987fc75c", "x3"}
            assert expected_ids == set(prepared_batches.groups.keys())
            for id_ in expected_ids:
                assert len(prepared_batches.get_group(id_).index) == 1

    @staticmethod
    def test_apply_revert_token_cols_wrapper() -> None:
        """Test `apply_revert_token_cols_wrapper`."""
        test_series = pd.Series(
            {
                "xID": 0,
                "xName": "test",
                "xSymbol": "T",
            }
        )
        res = apply_revert_token_cols_wrapper("x")(test_series)
        assert res == {
            "id": 0,
            "name": "test",
            "symbol": "T",
        }

    @staticmethod
    def test_revert_transform_hist_data(
        transformed_historical_data: pd.DataFrame,
        historical_data: Dict[str, List[Union[None, Dict[str, str], int, str, float]]],
    ) -> None:
        """Test `revert_transform_hist_data`."""
        reverted_pairs_hist = revert_transform_hist_data(transformed_historical_data)

        # Convert list of dicts into dict of lists.
        actual_reverted_pairs_hist: Dict[
            str, List[Union[None, Dict[str, str], int, str, float]]
        ] = {k: [] for k in reverted_pairs_hist[0].keys()}
        for (
            d
        ) in reverted_pairs_hist:  # you can list as many input dicts as you want here
            for key, value in d.items():
                actual_reverted_pairs_hist[key].append(value)

        assert actual_reverted_pairs_hist == historical_data
